"""Example Script: How to configure a SQL Datasource

This example script is intended for use in documentation on how to configure a SQL Datasource.

Assert statements are included to ensure that if the behaviour shown in this script breaks it will not pass
tests and will be updated.  These statements can be ignored by users.

Comments with the tags `<snippet>` and `</snippet>` are used to ensure that if this script is updated
the snippets that are specified for use in documentation are maintained.  These comments can be ignored by users.

--documentation--
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_sql_datasource

To run this code as a local test, use the following console command:
```
pytest -v --docs-tests -m integration -k "how_to_configure_a_sql_datasource" tests/integration/test_script_runner.py
```

To validate the snippets in this file, use the following console command:
```
yarn snippet-check ./tests/integration/docusaurus/connecting_to_your_data/datasource_configuration/how_to_configure_a_sql_datasource.py
```
"""
from ruamel import yaml

import great_expectations as gx
from tests.integration.docusaurus.connecting_to_your_data.datasource_configuration.datasource_configuration_test_utilities import (
    is_subset,
)
from tests.integration.docusaurus.connecting_to_your_data.datasource_configuration.full_datasource_configurations import (
    get_full_config_sql_configured_datasource,
    get_full_config_sql_inferred_datasource__single_and_multi_batch,
    get_full_config_sql_inferred_datasource__single_batch_only,
    get_full_config_sql_runtime_datasource,
    get_partial_config_universal_datasource_config_elements,
)

CONNECTION_STRING = "sqlite:///data/yellow_tripdata_sample_2020_all_months_combined.db"
data_context: gx.DataContext = gx.get_context()


def validate_universal_config_elements():
    """Validates that the 'universal' configuration keys and values are in fact identical to the keys and values
    in all the full Spark configurations.
    """
    universal_elements = get_partial_config_universal_datasource_config_elements()
    is_subset(
        universal_elements,
        get_full_config_sql_inferred_datasource__single_and_multi_batch(),
    )
    is_subset(
        universal_elements, get_full_config_sql_inferred_datasource__single_batch_only()
    )
    is_subset(universal_elements, get_full_config_sql_configured_datasource())
    is_subset(universal_elements, get_full_config_sql_runtime_datasource())


def section_5_add_the_sqlalchemy_execution_engine_to_your_datasource_configuration():
    datasource_config: dict = {
        # <snippet name="sql datasource define execution_engine class_name and module_name">
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        # </snippet>
    }
    for full_config in (
        get_full_config_sql_configured_datasource(),
        get_full_config_sql_inferred_datasource__single_batch_only(),
        get_full_config_sql_inferred_datasource__single_and_multi_batch(),
        get_full_config_sql_runtime_datasource(),
    ):
        is_subset(datasource_config, full_config)

    # <snippet name="sql datasource configuration post execution engine defined">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            # <snippet name="sql datasource define CONNECTION_STRING">
            "connection_string": CONNECTION_STRING,
            # </snippet>
        },
    }
    # </snippet>
    for full_config in (
        get_full_config_sql_configured_datasource(),
        get_full_config_sql_inferred_datasource__single_batch_only(),
        get_full_config_sql_inferred_datasource__single_and_multi_batch(),
        get_full_config_sql_runtime_datasource(),
    ):
        is_subset(datasource_config, full_config)


def section_6_add_a_dictionary_as_the_value_of_the_data_connectors_key():
    # <snippet name="sql datasource configuration post execution engine defined with empty data_connectors">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {},
    }
    # </snippet>
    for full_config in (
        get_full_config_sql_configured_datasource(),
        get_full_config_sql_inferred_datasource__single_batch_only(),
        get_full_config_sql_inferred_datasource__single_and_multi_batch(),
        get_full_config_sql_runtime_datasource(),
    ):
        is_subset(datasource_config, full_config)


def section_7_configure_your_individual_data_connectors__inferred():
    # <snippet name="sql datasource configuration with empty inferred sql data_connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {},
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_sql_inferred_datasource__single_batch_only()
    )

    # <snippet name="inferred sql datasource configuration with data_connector class_name defined">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="define data_connector class_name for inferred sql datasource">
                "class_name": "InferredAssetSqlDataConnector",
                # </snippet>
            },
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_sql_inferred_datasource__single_batch_only()
    )


def section_7_configure_your_individual_data_connectors__configured():
    # <snippet name="sql datasource configuration with empty configured sql data_connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {},
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_sql_configured_datasource())

    # <snippet name="configured sql datasource configuration with data_connector class_name defined">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                # <snippet name="define data_connector class_name for configured sql datasource">
                "class_name": "ConfiguredAssetSqlDataConnector",
                # </snippet>
                "assets": {},
            },
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_sql_configured_datasource())


def section_7_configure_your_individual_data_connectors__runtime():
    # <snippet name="sql datasource configuration with empty runtime sql data_connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {},
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_sql_runtime_datasource())

    # <snippet name="runtime sql datasource configuration with data_connector class_name defined">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                # <snippet name="define data_connector class_name for runtime sql datasource">
                "class_name": "RuntimeDataConnector",
                # </snippet>
                "batch_identifiers": {},
            },
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_sql_runtime_datasource())


def section_8_configure_your_data_connectors_data_assets__inferred():
    # <snippet name="full configuration for sql inferred Datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            # <snippet name="inferred sql data asset single batch">
            "inferred_data_connector_single_batch_asset": {
                "class_name": "InferredAssetSqlDataConnector",
            },
            # </snippet>
            # <snippet name="inferred sql data asset multi batch">
            "inferred_data_connector_multi_batch_asset_split_on_date_time": {
                "class_name": "InferredAssetSqlDataConnector",
                "splitter_method": "split_on_year_and_month",
                "splitter_kwargs": {
                    "column_name": "pickup_datetime",
                },
            },
            # </snippet>
        },
    }
    # </snippet>
    is_subset(
        datasource_config,
        get_full_config_sql_inferred_datasource__single_and_multi_batch(),
    )


def section_8_configure_your_data_connectors_data_assets__configured():
    # <snippet name="full configuration for sql configured Datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetSqlDataConnector",
                "assets": {
                    # <snippet name="configured sql data asset single batch">
                    "yellow_tripdata_sample_2020_full": {
                        "table_name": "yellow_tripdata_sample_2020",
                        "schema_name": "main",
                    },
                    # </snippet>
                    # <snippet name="configured sql data asset multi batch">
                    "yellow_tripdata_sample_2020_by_year_and_month": {
                        "table_name": "yellow_tripdata_sample_2020",
                        "schema_name": "main",
                        "splitter_method": "split_on_year_and_month",
                        "splitter_kwargs": {
                            "column_name": "pickup_datetime",
                        },
                    },
                    # </snippet>
                },
            },
        },
    }
    # </snippet>
    is_subset(
        datasource_config,
        get_full_config_sql_configured_datasource(),
    )


def section_8_configure_your_data_connectors_data_assets__runtime():
    # <snippet name="full configuration for sql runtime Datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                # <snippet name="runtime sql data asset define batch_identifiers">
                "batch_identifiers": ["batch_timestamp"],
                # </snippet>
            },
        },
    }
    # </snippet>
    is_subset(
        datasource_config,
        get_full_config_sql_runtime_datasource(),
    )


def configure_your_data_connectors_data_assets__runtime__yaml():
    # <snippet name="full configuration for sql runtime Datasource YAML">
    datasource_config = f"""
    name: my_datasource_name
    class_name: Datasource
    execution_engine:
      class_name: SqlAlchemyExecutionEngine
      connection_string: {CONNECTION_STRING}
    data_connectors:
      name_of_my_runtime_data_connector:
        class_name: RuntimeDataConnector
        batch_identifiers:
          - batch_timestamp
    """
    # </snippet>
    connector_name = "name_of_my_runtime_data_connector"
    asset_count = 0

    test_result = data_context.test_yaml_config(datasource_config)
    datasource_check = test_result.self_check(max_examples=12)
    assert (
        datasource_check["data_connectors"][connector_name]["data_asset_count"]
        == asset_count
    ), f"{connector_name} {asset_count} != {datasource_check['data_connectors'][connector_name]['data_asset_count']}"


def section_9_test_your_configuration__inferred_and_configured():
    for datasource_config, connector_name, asset_count, asset_name, batch_count in (
        (
            get_full_config_sql_inferred_datasource__single_and_multi_batch(),
            "inferred_data_connector_single_batch_asset",
            1,
            "yellow_tripdata_sample_2020",
            1,
        ),
        (
            get_full_config_sql_inferred_datasource__single_and_multi_batch(),
            "inferred_data_connector_multi_batch_asset_split_on_date_time",
            1,
            "yellow_tripdata_sample_2020",
            12,
        ),
        (
            get_full_config_sql_configured_datasource(),
            "name_of_my_configured_data_connector",
            2,
            "yellow_tripdata_sample_2020_full",
            1,
        ),
        (
            get_full_config_sql_configured_datasource(),
            "name_of_my_configured_data_connector",
            2,
            "yellow_tripdata_sample_2020_by_year_and_month",
            12,
        ),
    ):

        test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
        datasource_check = test_result.self_check(max_examples=12)
        assert (
            datasource_check["data_connectors"][connector_name]["data_asset_count"]
            == asset_count
        ), f"{connector_name} {asset_count} != {datasource_check['data_connectors'][connector_name]['data_asset_count']}"
        assert (
            datasource_check["data_connectors"][connector_name]["data_assets"][
                asset_name
            ]["batch_definition_count"]
            == batch_count
        ), f"{connector_name} {batch_count} batches != {datasource_check['data_connectors'][connector_name]['data_assets'][asset_name]['batch_definition_count']}"


def section_9_test_your_configuration__runtime():
    # Testing the runtime data connector (there will be no data assets, those are passed in by the RuntimeBatchRequest.)
    datasource_config = get_full_config_sql_runtime_datasource()
    connector_name = "name_of_my_runtime_data_connector"
    asset_count = 0

    test_result = (
        # <snippet name="test your sql datasource_config with test_yaml_config">
        data_context.test_yaml_config(yaml.dump(datasource_config))
        # </snippet>
    )
    datasource_check = test_result.self_check(max_examples=12)
    assert (
        datasource_check["data_connectors"][connector_name]["data_asset_count"]
        == asset_count
    ), f"{connector_name} {asset_count} != {datasource_check['data_connectors'][connector_name]['data_asset_count']}"


def run_tests():
    # Test to verify that the universal config elements are consistent in each of the SQL config examples.
    validate_universal_config_elements()

    # Test to verify that the sql specific config examples are consistent in each of the SQL config examples.
    section_5_add_the_sqlalchemy_execution_engine_to_your_datasource_configuration()
    section_6_add_a_dictionary_as_the_value_of_the_data_connectors_key()

    # Test to verify that the inferred config examples are consistent with each other.
    section_7_configure_your_individual_data_connectors__inferred()
    section_8_configure_your_data_connectors_data_assets__inferred()

    # Test to verify that the configured config examples are consistent with each other.
    section_7_configure_your_individual_data_connectors__configured()
    section_8_configure_your_data_connectors_data_assets__configured()

    # Test to verify that the runtime config examples are consistent with each other.
    section_7_configure_your_individual_data_connectors__runtime()
    section_8_configure_your_data_connectors_data_assets__runtime()
    configure_your_data_connectors_data_assets__runtime__yaml()

    # Test to verify that the full configuration examples are functional.
    section_9_test_your_configuration__inferred_and_configured()
    section_9_test_your_configuration__runtime()


run_tests()



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, List, Optional, Set, Union

import numpy as np

from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.core.metric_function_types import (
    SummarizationMetricNameSuffixes,
)
from great_expectations.rule_based_profiler.config import (
    ParameterBuilderConfig,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.helpers.util import (
    NP_EPSILON,
    get_parameter_value_and_validate_return_type,
)
from great_expectations.rule_based_profiler.metric_computation_result import (
    MetricValues,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.parameter_builder import (
    MetricMultiBatchParameterBuilder,
)
from great_expectations.rule_based_profiler.parameter_container import (
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    RAW_PARAMETER_KEY,
    ParameterContainer,
    ParameterNode,
)
from great_expectations.types.attributes import Attributes

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class MeanUnexpectedMapMetricMultiBatchParameterBuilder(
    MetricMultiBatchParameterBuilder
):
    """
    Compute mean unexpected count ratio (as a fraction) of specified map-style metric across every Batch of data given.
    """

    exclude_field_names: ClassVar[
        Set[str]
    ] = MetricMultiBatchParameterBuilder.exclude_field_names | {
        "metric_name",
        "single_batch_mode",
        "enforce_numeric_metric",
        "replace_nan_with_zero",
        "reduce_scalar_metric",
    }

    def __init__(
        self,
        name: str,
        map_metric_name: str,
        total_count_parameter_builder_name: str,
        null_count_parameter_builder_name: Optional[str] = None,
        metric_domain_kwargs: Optional[Union[str, dict]] = None,
        metric_value_kwargs: Optional[Union[str, dict]] = None,
        evaluation_parameter_builder_configs: Optional[
            List[ParameterBuilderConfig]
        ] = None,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        """
        Args:
            name: the name of this parameter -- this is user-specified parameter name (from configuration);
            it is not the fully-qualified parameter name; a fully-qualified parameter name must start with "$parameter."
            and may contain one or more subsequent parts (e.g., "$parameter.<my_param_from_config>.<metric_name>").
            map_metric_name: the name of a map metric (must be a supported and registered map metric); the suffix
            ".unexpected_count" will be appended to "map_metric_name" to be used in MetricConfiguration to get values.
            total_count_parameter_builder_name: name of parameter that computes total_count (of rows in Batch).
            null_count_parameter_builder_name: name of parameter that computes null_count (of domain values in Batch).
            metric_domain_kwargs: used in MetricConfiguration
            metric_value_kwargs: used in MetricConfiguration
            evaluation_parameter_builder_configs: ParameterBuilder configurations, executing and making whose respective
            ParameterBuilder objects' outputs available (as fully-qualified parameter names) is pre-requisite.
            These "ParameterBuilder" configurations help build parameters needed for this "ParameterBuilder".
            data_context: AbstractDataContext associated with this ParameterBuilder
        """
        super().__init__(
            name=name,
            metric_name=f"{map_metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
            metric_domain_kwargs=metric_domain_kwargs,
            metric_value_kwargs=metric_value_kwargs,
            enforce_numeric_metric=True,
            replace_nan_with_zero=True,
            reduce_scalar_metric=True,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )

        self._map_metric_name = map_metric_name
        self._total_count_parameter_builder_name = total_count_parameter_builder_name
        self._null_count_parameter_builder_name = null_count_parameter_builder_name

    @property
    def map_metric_name(self) -> str:
        return self._map_metric_name

    @property
    def total_count_parameter_builder_name(self) -> str:
        return self._total_count_parameter_builder_name

    @property
    def null_count_parameter_builder_name(self) -> Optional[str]:
        return self._null_count_parameter_builder_name

    def _build_parameters(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> Attributes:
        """
        Builds ParameterContainer object that holds ParameterNode objects with attribute name-value pairs and details.

        Returns:
            Attributes object, containing computed parameter values and parameter computation details metadata.
        """
        # Obtain total_count_parameter_builder_name from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        total_count_parameter_builder_name: str = (
            get_parameter_value_and_validate_return_type(
                domain=domain,
                parameter_reference=self.total_count_parameter_builder_name,
                expected_return_type=str,
                variables=variables,
                parameters=parameters,
            )
        )

        fully_qualified_total_count_parameter_builder_name: str = (
            f"{RAW_PARAMETER_KEY}{total_count_parameter_builder_name}"
        )
        # Obtain total_count from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        total_count_parameter_node: ParameterNode = (
            get_parameter_value_and_validate_return_type(
                domain=domain,
                parameter_reference=fully_qualified_total_count_parameter_builder_name,
                expected_return_type=None,
                variables=variables,
                parameters=parameters,
            )
        )
        total_count_values: MetricValues = total_count_parameter_node[
            FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY
        ]

        # Obtain null_count_parameter_builder_name from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        null_count_parameter_builder_name: Optional[
            str
        ] = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.null_count_parameter_builder_name,
            expected_return_type=None,
            variables=variables,
            parameters=parameters,
        )

        batch_ids: Optional[List[str]] = self.get_batch_ids(
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        num_batch_ids: int = len(batch_ids)

        null_count_values: MetricValues
        if null_count_parameter_builder_name is None:
            null_count_values = np.zeros(shape=(num_batch_ids,))
        else:
            fully_qualified_null_count_parameter_builder_name: str = (
                f"{RAW_PARAMETER_KEY}{null_count_parameter_builder_name}"
            )
            # Obtain null_count from "rule state" (i.e., variables and parameters); from instance variable otherwise.
            null_count_parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
                domain=domain,
                parameter_reference=fully_qualified_null_count_parameter_builder_name,
                expected_return_type=None,
                variables=variables,
                parameters=parameters,
            )
            null_count_values = null_count_parameter_node[
                FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY
            ]

        nonnull_count_values: np.ndarray = total_count_values - null_count_values

        # Compute "unexpected_count" corresponding to "map_metric_name" (given as argument to this "ParameterBuilder").
        super().build_parameters(
            domain=domain,
            variables=variables,
            parameters=parameters,
            parameter_computation_impl=super()._build_parameters,
            runtime_configuration=runtime_configuration,
        )

        # Retrieve "unexpected_count" corresponding to "map_metric_name" (given as argument to this "ParameterBuilder").
        parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.raw_fully_qualified_parameter_name,
            expected_return_type=None,
            variables=variables,
            parameters=parameters,
        )
        unexpected_count_values: MetricValues = parameter_node[
            FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY
        ]

        unexpected_count_ratio_values: np.ndarray = unexpected_count_values / (
            nonnull_count_values + NP_EPSILON
        )
        mean_unexpected_count_ratio: np.float64 = np.mean(unexpected_count_ratio_values)

        return Attributes(
            {
                FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY: mean_unexpected_count_ratio,
                FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY: parameter_node[
                    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY
                ],
            }
        )



from __future__ import annotations

import logging
import os
import pathlib
import shutil
import warnings
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from ruamel.yaml import YAML

import great_expectations.exceptions as gx_exceptions
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.templates import (
    CONFIG_VARIABLES_TEMPLATE,
    PROJECT_TEMPLATE_USAGE_STATISTICS_DISABLED,
    PROJECT_TEMPLATE_USAGE_STATISTICS_ENABLED,
)
from great_expectations.data_context.types.base import (
    CURRENT_GX_CONFIG_VERSION,
    MINIMUM_SUPPORTED_CONFIG_VERSION,
    AnonymizedUsageStatisticsConfig,
    DataContextConfig,
    DataContextConfigDefaults,
)
from great_expectations.data_context.util import file_relative_path

logger = logging.getLogger(__name__)
yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False

if TYPE_CHECKING:
    from great_expectations.alias_types import JSONValues, PathStr


class SerializableDataContext(AbstractDataContext):
    """
    TODO - write docstring!
    """

    UNCOMMITTED_DIRECTORIES = ["data_docs", "validations"]
    GX_UNCOMMITTED_DIR = "uncommitted"
    BASE_DIRECTORIES = [
        DataContextConfigDefaults.CHECKPOINTS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.EXPECTATIONS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.PLUGINS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.PROFILERS_BASE_DIRECTORY.value,
        GX_UNCOMMITTED_DIR,
    ]
    GX_DIR: ClassVar[str] = "great_expectations"
    GX_YML: ClassVar[str] = "great_expectations.yml"
    GX_EDIT_NOTEBOOK_DIR = GX_UNCOMMITTED_DIR
    DOLLAR_SIGN_ESCAPE_STRING = r"\$"

    def __init__(
        self,
        context_root_dir: PathStr,
        runtime_environment: Optional[dict] = None,
    ) -> None:
        if isinstance(context_root_dir, pathlib.Path):
            # TODO: (kilo59) 122022 should be saving and passing around `pathlib.Path` not str
            context_root_dir = str(context_root_dir)
        self._context_root_directory = context_root_dir
        super().__init__(runtime_environment=runtime_environment)

    def _init_datasource_store(self):
        raise NotImplementedError  # Required by parent ABC but this class is never instantiated

    def _init_variables(self):
        raise NotImplementedError  # Required by parent ABC but this class is never instantiated

    def _save_project_config(self) -> None:
        """
        See parent 'AbstractDataContext._save_project_config()` for more information.

        Explicitly override base class implementation to retain legacy behavior.
        """
        config_filepath = os.path.join(self.root_directory, self.GX_YML)  # type: ignore[arg-type]

        logger.debug(
            f"Starting DataContext._save_project_config; attempting to update {config_filepath}"
        )

        try:
            with open(config_filepath, "w") as outfile:

                zep_datasources = self._synchronize_zep_datasources()
                if zep_datasources:
                    self.zep_config.datasources.update(zep_datasources)
                    logger.info(
                        f"Saving {len(self.zep_config.datasources)} ZEP Datasources to {config_filepath}"
                    )
                    zep_json_dict: dict[str, JSONValues] = self.zep_config._json_dict()
                    self.config._commented_map.update(zep_json_dict)

                self.config.to_yaml(outfile)
        except PermissionError as e:
            logger.warning(f"Could not save project config to disk: {e}")

    def _check_for_usage_stats_sync(self, project_config: DataContextConfig) -> bool:
        """
        If there are differences between the DataContextConfig used to instantiate
        the DataContext and the DataContextConfig assigned to `self.config`, we want
        to save those changes to disk so that subsequent instantiations will utilize
        the same values.

        A small caveat is that if that difference stems from a global override (env var
        or conf file), we don't want to write to disk. This is due to the fact that
        those mechanisms allow for dynamic values and saving them will make them static.

        Args:
            project_config: The DataContextConfig used to instantiate the DataContext.

        Returns:
            A boolean signifying whether or not the current DataContext's config needs
            to be persisted in order to recognize changes made to usage statistics.
        """
        project_config_usage_stats: Optional[
            AnonymizedUsageStatisticsConfig
        ] = project_config.anonymous_usage_statistics
        context_config_usage_stats: Optional[
            AnonymizedUsageStatisticsConfig
        ] = self.config.anonymous_usage_statistics

        if (
            project_config_usage_stats.enabled is False  # type: ignore[union-attr]
            or context_config_usage_stats.enabled is False  # type: ignore[union-attr]
        ):
            return False

        if project_config_usage_stats.explicit_id is False:  # type: ignore[union-attr]
            return True

        if project_config_usage_stats == context_config_usage_stats:
            return False

        if project_config_usage_stats is None or context_config_usage_stats is None:
            return True

        # If the data_context_id differs and that difference is not a result of a global override, a sync is necessary.
        global_data_context_id: Optional[str] = self._get_data_context_id_override()
        if (
            project_config_usage_stats.data_context_id
            != context_config_usage_stats.data_context_id
            and context_config_usage_stats.data_context_id != global_data_context_id
        ):
            return True

        # If the usage_statistics_url differs and that difference is not a result of a global override, a sync is necessary.
        global_usage_stats_url: Optional[str] = self._get_usage_stats_url_override()
        if (
            project_config_usage_stats.usage_statistics_url
            != context_config_usage_stats.usage_statistics_url
            and context_config_usage_stats.usage_statistics_url
            != global_usage_stats_url
        ):
            return True

        return False

    @classmethod
    def create(
        cls,
        project_root_dir: Optional[PathStr] = None,
        usage_statistics_enabled: bool = True,
        runtime_environment: Optional[dict] = None,
    ) -> SerializableDataContext:
        """
        Build a new great_expectations directory and DataContext object in the provided project_root_dir.

        `create` will create a new "great_expectations" directory in the provided folder, provided one does not
        already exist. Then, it will initialize a new DataContext in that folder and write the resulting config.

        --Public API--

        --Documentation--
            https://docs.greatexpectations.io/docs/terms/data_context

        Args:
            project_root_dir: path to the root directory in which to create a new great_expectations directory
            usage_statistics_enabled: boolean directive specifying whether or not to gather usage statistics
            runtime_environment: a dictionary of config variables that override both those set in
                config_variables.yml and the environment

        Returns:
            DataContext
        """

        if not os.path.isdir(project_root_dir):  # type: ignore[arg-type]
            raise gx_exceptions.DataContextError(
                "The project_root_dir must be an existing directory in which "
                "to initialize a new DataContext"
            )

        gx_dir = os.path.join(project_root_dir, cls.GX_DIR)  # type: ignore[arg-type]
        os.makedirs(gx_dir, exist_ok=True)
        cls._scaffold_directories(gx_dir)

        if os.path.isfile(os.path.join(gx_dir, cls.GX_YML)):
            message = f"""Warning. An existing `{cls.GX_YML}` was found here: {gx_dir}.
    - No action was taken."""
            warnings.warn(message)
        else:
            cls._write_project_template_to_disk(gx_dir, usage_statistics_enabled)

        uncommitted_dir = os.path.join(gx_dir, cls.GX_UNCOMMITTED_DIR)
        if os.path.isfile(os.path.join(uncommitted_dir, "config_variables.yml")):
            message = """Warning. An existing `config_variables.yml` was found here: {}.
    - No action was taken.""".format(
                uncommitted_dir
            )
            warnings.warn(message)
        else:
            cls._write_config_variables_template_to_disk(uncommitted_dir)

        return cls(context_root_dir=gx_dir, runtime_environment=runtime_environment)

    @classmethod
    def all_uncommitted_directories_exist(cls, gx_dir: PathStr) -> bool:
        """Check if all uncommitted directories exist."""
        uncommitted_dir = os.path.join(gx_dir, cls.GX_UNCOMMITTED_DIR)
        for directory in cls.UNCOMMITTED_DIRECTORIES:
            if not os.path.isdir(os.path.join(uncommitted_dir, directory)):
                return False

        return True

    @classmethod
    def config_variables_yml_exist(cls, gx_dir: PathStr) -> bool:
        """Check if all config_variables.yml exists."""
        path_to_yml = os.path.join(gx_dir, cls.GX_YML)

        # TODO this is so brittle and gross
        with open(path_to_yml) as f:
            config = yaml.load(f)
        config_var_path = config.get("config_variables_file_path")
        config_var_path = os.path.join(gx_dir, config_var_path)
        return os.path.isfile(config_var_path)

    @classmethod
    def _write_config_variables_template_to_disk(cls, uncommitted_dir: str) -> None:
        os.makedirs(uncommitted_dir, exist_ok=True)
        config_var_file = os.path.join(uncommitted_dir, "config_variables.yml")
        with open(config_var_file, "w") as template:
            template.write(CONFIG_VARIABLES_TEMPLATE)

    @classmethod
    def _write_project_template_to_disk(
        cls, gx_dir: PathStr, usage_statistics_enabled: bool = True
    ) -> None:
        file_path = os.path.join(gx_dir, cls.GX_YML)
        with open(file_path, "w") as template:
            if usage_statistics_enabled:
                template.write(PROJECT_TEMPLATE_USAGE_STATISTICS_ENABLED)
            else:
                template.write(PROJECT_TEMPLATE_USAGE_STATISTICS_DISABLED)

    @classmethod
    def _scaffold_directories(cls, base_dir: PathStr) -> None:
        """Safely create GE directories for a new project."""
        os.makedirs(base_dir, exist_ok=True)
        with open(os.path.join(base_dir, ".gitignore"), "w") as f:
            f.write("uncommitted/")

        for directory in cls.BASE_DIRECTORIES:
            if directory == "plugins":
                plugins_dir = os.path.join(base_dir, directory)
                os.makedirs(plugins_dir, exist_ok=True)
                os.makedirs(
                    os.path.join(plugins_dir, "custom_data_docs"), exist_ok=True
                )
                os.makedirs(
                    os.path.join(plugins_dir, "custom_data_docs", "views"),
                    exist_ok=True,
                )
                os.makedirs(
                    os.path.join(plugins_dir, "custom_data_docs", "renderers"),
                    exist_ok=True,
                )
                os.makedirs(
                    os.path.join(plugins_dir, "custom_data_docs", "styles"),
                    exist_ok=True,
                )
                cls._scaffold_custom_data_docs(plugins_dir)
            else:
                os.makedirs(os.path.join(base_dir, directory), exist_ok=True)

        uncommitted_dir = os.path.join(base_dir, cls.GX_UNCOMMITTED_DIR)

        for new_directory in cls.UNCOMMITTED_DIRECTORIES:
            new_directory_path = os.path.join(uncommitted_dir, new_directory)
            os.makedirs(new_directory_path, exist_ok=True)

    @classmethod
    def _scaffold_custom_data_docs(cls, plugins_dir: PathStr) -> None:
        """Copy custom data docs templates"""
        styles_template = file_relative_path(
            __file__,
            "../../render/view/static/styles/data_docs_custom_styles_template.css",
        )
        styles_destination_path = os.path.join(
            plugins_dir, "custom_data_docs", "styles", "data_docs_custom_styles.css"
        )
        shutil.copyfile(styles_template, styles_destination_path)

    @classmethod
    def find_context_root_dir(cls) -> str:
        """
        TODO
        """
        result = None
        yml_path = None
        gx_home_environment = os.getenv("GX_HOME")
        if gx_home_environment:
            gx_home_environment = os.path.expanduser(gx_home_environment)
            if os.path.isdir(gx_home_environment) and os.path.isfile(
                os.path.join(gx_home_environment, "great_expectations.yml")
            ):
                result = gx_home_environment
        else:
            yml_path = cls._find_context_yml_file()
            if yml_path:
                result = os.path.dirname(yml_path)

        if result is None:
            raise gx_exceptions.ConfigNotFoundError()

        logger.debug(f"Using project config: {yml_path}")
        return result

    @classmethod
    def get_ge_config_version(
        cls, context_root_dir: Optional[PathStr] = None
    ) -> Optional[float]:
        """
        TODO
        """
        yml_path = cls._find_context_yml_file(search_start_dir=context_root_dir)
        if yml_path is None:
            return None

        with open(yml_path) as f:
            config_commented_map_from_yaml = yaml.load(f)

        config_version = config_commented_map_from_yaml.get("config_version")
        return float(config_version) if config_version else None

    @classmethod
    def set_ge_config_version(
        cls,
        config_version: Union[int, float],
        context_root_dir: Optional[str] = None,
        validate_config_version: bool = True,
    ) -> bool:
        """
        TODO
        """
        if not isinstance(config_version, (int, float)):
            raise gx_exceptions.UnsupportedConfigVersionError(
                "The argument `config_version` must be a number.",
            )

        if validate_config_version:
            if config_version < MINIMUM_SUPPORTED_CONFIG_VERSION:
                raise gx_exceptions.UnsupportedConfigVersionError(
                    "Invalid config version ({}).\n    The version number must be at least {}. ".format(
                        config_version, MINIMUM_SUPPORTED_CONFIG_VERSION
                    ),
                )
            elif config_version > CURRENT_GX_CONFIG_VERSION:
                raise gx_exceptions.UnsupportedConfigVersionError(
                    "Invalid config version ({}).\n    The maximum valid version is {}.".format(
                        config_version, CURRENT_GX_CONFIG_VERSION
                    ),
                )

        yml_path = cls._find_context_yml_file(search_start_dir=context_root_dir)
        if yml_path is None:
            return False

        with open(yml_path) as f:
            config_commented_map_from_yaml = yaml.load(f)
            config_commented_map_from_yaml["config_version"] = float(config_version)

        with open(yml_path, "w") as f:
            yaml.dump(config_commented_map_from_yaml, f)

        return True

    @classmethod
    def _find_context_yml_file(
        cls, search_start_dir: Optional[PathStr] = None
    ) -> Optional[str]:
        """Search for the yml file starting here and moving upward."""
        yml_path = None
        if search_start_dir is None:
            search_start_dir = os.getcwd()

        for i in range(4):
            logger.debug(
                f"Searching for config file {search_start_dir} ({i} layer deep)"
            )

            potential_ge_dir = os.path.join(search_start_dir, cls.GX_DIR)

            if os.path.isdir(potential_ge_dir):
                potential_yml = os.path.join(potential_ge_dir, cls.GX_YML)
                if os.path.isfile(potential_yml):
                    yml_path = potential_yml
                    logger.debug(f"Found config file at {str(yml_path)}")
                    break
            # move up one directory
            search_start_dir = os.path.dirname(search_start_dir)

        return yml_path

    @classmethod
    def does_config_exist_on_disk(cls, context_root_dir: PathStr) -> bool:
        """Return True if the great_expectations.yml exists on disk."""
        return os.path.isfile(os.path.join(context_root_dir, cls.GX_YML))

    @classmethod
    def is_project_initialized(cls, ge_dir: PathStr) -> bool:
        """
        Return True if the project is initialized.

        To be considered initialized, all of the following must be true:
        - all project directories exist (including uncommitted directories)
        - a valid great_expectations.yml is on disk
        - a config_variables.yml is on disk
        - the project has at least one datasource
        - the project has at least one suite
        """
        return (
            cls.does_config_exist_on_disk(ge_dir)
            and cls.all_uncommitted_directories_exist(ge_dir)
            and cls.config_variables_yml_exist(ge_dir)
            and cls._does_context_have_at_least_one_datasource(ge_dir)
            and cls._does_context_have_at_least_one_suite(ge_dir)
        )

    @classmethod
    def _does_project_have_a_datasource_in_config_file(cls, ge_dir: PathStr) -> bool:
        if not cls.does_config_exist_on_disk(ge_dir):
            return False
        return cls._does_context_have_at_least_one_datasource(ge_dir)

    @classmethod
    def _does_context_have_at_least_one_datasource(cls, ge_dir: PathStr) -> bool:
        context = cls._attempt_context_instantiation(ge_dir)
        if not context:
            return False
        return len(context.list_datasources()) >= 1

    @classmethod
    def _does_context_have_at_least_one_suite(cls, ge_dir: PathStr) -> bool:
        context = cls._attempt_context_instantiation(ge_dir)
        if not context:
            return False
        return bool(context.list_expectation_suites())

    @classmethod
    def _attempt_context_instantiation(
        cls, ge_dir: PathStr
    ) -> Optional[SerializableDataContext]:
        try:
            context = cls(context_root_dir=ge_dir)
            return context
        except (
            gx_exceptions.DataContextError,
            gx_exceptions.InvalidDataContextConfigError,
        ) as e:
            logger.debug(e)
        return None



from typing import List

import numpy as np

from great_expectations.core.batch import BatchRequest
from great_expectations.data_context.data_context import DataContext
from great_expectations.datasource.data_connector.batch_filter import (
    BatchFilter,
    build_batch_filter,
)
from great_expectations.validator.metric_configuration import MetricConfiguration
from great_expectations.validator.validator import Validator

context = DataContext()
suite = context.get_expectation_suite("yellow_tripdata_validations")

# Create three BatchRequests for Jan, Feb, and March 2019 data and instantiate a Validator with all three BatchRequests
jan_batch_request: BatchRequest = BatchRequest(
    datasource_name="taxi_pandas",
    data_connector_name="monthly",
    data_asset_name="my_reports",
    data_connector_query={"batch_filter_parameters": {"month": "01", "year": "2019"}},
)

feb_batch_request: BatchRequest = BatchRequest(
    datasource_name="taxi_pandas",
    data_connector_name="monthly",
    data_asset_name="my_reports",
    data_connector_query={"batch_filter_parameters": {"month": "02", "year": "2019"}},
)

march_batch_request: BatchRequest = BatchRequest(
    datasource_name="taxi_pandas",
    data_connector_name="monthly",
    data_asset_name="my_reports",
    data_connector_query={"batch_filter_parameters": {"month": "03", "year": "2019"}},
)

validator: Validator = context.get_validator(
    batch_request_list=[jan_batch_request, feb_batch_request, march_batch_request],
    expectation_suite=suite,
)
assert validator.active_batch_definition.batch_identifiers["month"] == "03"
assert validator.active_batch_definition.batch_identifiers["year"] == "2019"

# Get the list of all batches contained by the Validator for use in the BatchFileter
total_batch_definition_list: List = [
    v.batch_definition for k, v in validator.batches.items()
]

# Filter to all batch_definitions prior to March
jan_feb_batch_filter: BatchFilter = build_batch_filter(
    data_connector_query_dict={
        "custom_filter_function": lambda batch_identifiers: int(
            batch_identifiers["month"]
        )
        < 3
    }
)
jan_feb_batch_definition_list: list = (
    jan_feb_batch_filter.select_from_data_connector_query(
        batch_definition_list=total_batch_definition_list
    )
)

# Get the highest max and lowest min between January and February
cumulative_max = 0
cumulative_min = np.Inf
for batch_definition in jan_feb_batch_definition_list:
    batch_id: str = batch_definition.id
    current_max = validator.get_metric(
        MetricConfiguration(
            "column.max",
            metric_domain_kwargs={"column": "fare_amount", "batch_id": batch_id},
        )
    )
    cumulative_max = current_max if current_max > cumulative_max else cumulative_max

    current_min = validator.get_metric(
        MetricConfiguration(
            "column.min",
            metric_domain_kwargs={"column": "fare_amount", "batch_id": batch_id},
        )
    )
    cumulative_min = current_min if current_min < cumulative_min else cumulative_min

# Use the highest max and lowest min from Jan and Feb to create an expectation which we validate against March
result = validator.expect_column_values_to_be_between(
    "fare_amount", min_value=cumulative_min, max_value=cumulative_max
)
assert result["success"]



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

from great_expectations.types import SerializableDictDot


class TestData(dict):
    __test__ = False  # Tell pytest not to try to collect this class as a test
    pass


class Backend(Enum):
    """Backends with some level of testing and support"""

    BIGQUERY = "CONCEPT_ONLY"
    MSSQL = "EXPERIMENTAL"
    SQLITE = "BETA"
    PYSPARK = "PRODUCTION"


@dataclass
class TestBackend:
    backend: str
    dialects: List[str]

    __test__ = False  # Tell pytest not to try to collect this class as a test

    def __post_init__(self) -> None:
        allowed_backend_names = ("pandas", "spark", "sqlalchemy")
        allowed_sql_dialects = ("sqlite", "postgresql", "mysql", "mssql", "bigquery")
        assert (
            self.backend in allowed_backend_names
        ), f"backend must be one of {allowed_backend_names}, not {self.backend}"
        if self.backend != "sqlalchemy":
            assert (
                self.dialects is None
            ), f"You may not specify dialects for backend {self.backend}"
        else:
            assert (
                type(self.dialects) == list and len(self.dialects) > 0
            ), "dialects must be a list for backend sqlalchemy"
            bad_dialects = [
                dialect
                for dialect in self.dialects
                if dialect not in allowed_sql_dialects
            ]
            assert (
                bad_dialects == []
            ), f"dialects can only include {allowed_sql_dialects}, not {bad_dialects}"


@dataclass
class ExpectationTestCase(SerializableDictDot):
    """A single test case, with input arguments and output"""

    title: str
    input: Dict[str, Any]
    output: Dict[str, Any]
    exact_match_out: bool
    suppress_test_for: List[str] = field(default_factory=list)
    include_in_gallery: bool = False
    only_for: Optional[List[str]] = None


class ExpectationLegacyTestCaseAdapter(ExpectationTestCase):
    """This class provides an adapter between the test cases developed prior to Great Expectations' 0.14 release and the newer ExpectationTestCase dataclass

    Notes:
    * Legacy test cases used "in" (a python reserved word). This has been changed to "input".
    * To maintain parallelism, we've also made the corresponding change from "out" to "output".
    * To avoid any ambiguity, ExpectationLegacyTestCaseAdapter only accepts keyword arguments. Positional arguments are not allowed.
    """

    def __init__(
        self,
        *,
        title,
        exact_match_out,
        out,
        suppress_test_for=[],
        **kwargs,
    ) -> None:
        super().__init__(
            title=title,
            input=kwargs["in"],
            output=out,
            exact_match_out=exact_match_out,
            suppress_test_for=suppress_test_for,
            only_for=kwargs.get("only_for"),
            include_in_gallery=kwargs.get("include_in_gallery", False),
        )


@dataclass
class ExpectationTestDataCases(SerializableDictDot):
    """Pairs a test dataset and a list of test cases to execute against that data."""

    data: TestData
    tests: List[ExpectationTestCase]
    schemas: Dict[Backend, Dict[str, str]] = field(default_factory=dict)
    test_backends: Optional[List[TestBackend]] = None
    data_alt: Optional[TestData] = None



from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING, Optional

from great_expectations.datasource.data_connector.configured_asset_sql_data_connector import (
    ConfiguredAssetSqlDataConnector,  # noqa: TCH001
)
from great_expectations.datasource.new_datasource import BaseDatasource

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine as SaEngine

    from great_expectations.execution_engine import SqlAlchemyExecutionEngine

logger = logging.getLogger(__name__)


class SimpleSqlalchemyDatasource(BaseDatasource):
    """A specialized Datasource for SQL backends

    SimpleSqlalchemyDatasource is designed to minimize boilerplate configuration and new concepts

    # <Alex>8/25/2022</Alex>
    Formal "SQLAlchemy" "DataConnector" classes, "ConfiguredAssetSqlDataConnector" and "InferredAssetSqlDataConnector",
    have been certified; hence, they should be used going forward.  On the other hand, "SimpleSqlalchemyDatasource",
    fully implemented on top of "ConfiguredAssetSqlDataConnector" and "InferredAssetSqlDataConnector", remains
    potentially useful for demonstration purposes, and should be maintained, even though it violates abstractions.
    """

    execution_engine: SqlAlchemyExecutionEngine

    def __init__(
        self,
        name: str,
        connection_string: Optional[str] = None,
        url: Optional[str] = None,
        credentials: Optional[dict] = None,
        engine: Optional[SaEngine] = None,  # sqlalchemy.engine.Engine
        introspection: Optional[dict] = None,
        tables: Optional[dict] = None,
        **kwargs,
    ) -> None:
        introspection = introspection or {}
        tables = tables or {}

        # In "SQLAlchemy" datasources, "connection_string" resides in "SqlAlchemyExecutionEngine" (not "DataConnector").
        self._execution_engine_config = {
            "class_name": "SqlAlchemyExecutionEngine",
            "connection_string": connection_string,
            "url": url,
            "credentials": credentials,
            "engine": engine,
        }
        self._execution_engine_config.update(**kwargs)

        super().__init__(name=name, execution_engine=self._execution_engine_config)

        self._data_connectors: dict = {}

        self._init_data_connectors(
            introspection_configs=introspection,
            table_configs=tables,
        )

        # NOTE: Abe 20201111 : This is incorrect. Will need to be fixed when we reconcile all the configs.
        self._datasource_config: dict = {}

    # noinspection PyMethodOverriding
    # Note: This method is meant to overwrite Datasource._init_data_connectors (despite signature mismatch).
    def _init_data_connectors(
        self,
        introspection_configs: dict,
        table_configs: dict,
    ) -> None:
        # Step-1: Build "DataConnector" objects for introspected tables/views (using "InferredAssetSqlDataConnector").
        for name, config in introspection_configs.items():
            data_connector_config = dict(
                **{
                    "class_name": "InferredAssetSqlDataConnector",
                    "name": name,
                },
                **config,
            )
            self._build_data_connector_from_config(
                name=name,
                config=data_connector_config,
            )

        # Step-2: Build "DataConnector" objects for tables (using "ConfiguredAssetSqlDataConnector").
        for data_asset_name, table_config in table_configs.items():
            for partitioner_name, partitioner_config in table_config[
                "partitioners"
            ].items():
                data_connector_name: str = partitioner_name
                if data_connector_name not in self.data_connectors:
                    data_connector_config = {
                        "class_name": "ConfiguredAssetSqlDataConnector",
                        "assets": {},
                    }
                    self._build_data_connector_from_config(
                        name=data_connector_name, config=data_connector_config
                    )

                data_connector: ConfiguredAssetSqlDataConnector = self.data_connectors[
                    data_connector_name
                ]

                data_asset_config: dict = copy.deepcopy(partitioner_config)

                if "table_name" not in data_asset_config:
                    data_asset_config["table_name"] = data_asset_name

                # "ConfiguredAssetSqlDataConnector" has only one asset type: "table" (adding for interface consistency).
                data_asset_config["type"] = "table"

                # If config contains any prefix, suffix or schema_name values,
                # they will be handled at the ConfiguredAssetSqlDataConnector-level
                data_connector.add_data_asset(
                    name=data_asset_name,
                    config=data_asset_config,
                )



import pytest

from tests.test_utils import get_awsathena_connection_url


@pytest.mark.unit
def test_get_awsathena_connection_url(monkeypatch):
    monkeypatch.setenv("ATHENA_STAGING_S3", "s3://test-staging/")
    monkeypatch.setenv("ATHENA_DB_NAME", "test_db_name")
    monkeypatch.setenv("ATHENA_TEN_TRIPS_DB_NAME", "test_ten_trips_db_name")

    assert (
        get_awsathena_connection_url()
        == "awsathena+rest://@athena.us-east-1.amazonaws.com/test_db_name?s3_staging_dir=s3://test-staging/"
    )

    assert (
        get_awsathena_connection_url(db_name_env_var="ATHENA_TEN_TRIPS_DB_NAME")
        == "awsathena+rest://@athena.us-east-1.amazonaws.com/test_ten_trips_db_name?s3_staging_dir=s3://test-staging/"
    )



"""
This is a basic generated Great Expectations script that runs a Checkpoint.

Checkpoints are the primary method for validating batches of data in production and triggering any followup actions.

A Checkpoint facilitates running a validation as well as configurable Actions such as updating Data Docs, sending a
notification to team members about validation results, or storing a result in a shared cloud storage.

See also <cyan>https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_create_a_new_checkpoint_using_test_yaml_config.html</cyan> for more information about the Checkpoints and how to configure them in your Great Expectations environment.

Checkpoints can be run directly without this script using the `great_expectations checkpoint run` command.  This script
is provided for those who wish to run Checkpoints in python.

Usage:
- Run this file: `python great_expectations/uncommitted/run_{0:s}.py`.
- This can be run manually or via a scheduler such, as cron.
- If your pipeline runner supports python snippets, then you can paste this into your pipeline.
"""
import sys

from great_expectations.checkpoint.types.checkpoint_result import (
    CheckpointResult,  # noqa: TCH001
)
from great_expectations.data_context import FileDataContext  # noqa: TCH001
from great_expectations.util import get_context

data_context: FileDataContext = get_context(context_root_dir="{1:s}")

result: CheckpointResult = data_context.run_checkpoint(
    checkpoint_name="{0:s}",
    batch_request=None,
    run_name=None,
)

if not result["success"]:
    print("Validation failed!")
    sys.exit(1)

print("Validation succeeded!")
sys.exit(0)



from typing import Any, Dict, List, Union

import pytest

try:
    sqlalchemy = pytest.importorskip("sqlalchemy")
except ImportError:
    sqlalchemy = None

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch import Batch, RuntimeBatchRequest
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource.new_datasource import Datasource

yaml = YAMLHandler()


@pytest.fixture
def datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine(db_file, sa):
    basic_datasource: Datasource = instantiate_class_from_config(
        yaml.load(
            f"""
    class_name: Datasource

    execution_engine:
        class_name: SqlAlchemyExecutionEngine
        connection_string: sqlite:///{db_file}

    data_connectors:
        test_runtime_data_connector:
            module_name: great_expectations.datasource.data_connector
            class_name: RuntimeDataConnector
            batch_identifiers:
                - pipeline_stage_name
                - airflow_run_id
                - custom_key_0
            assets:
                asset_a:
                    batch_identifiers:
                        - day
                        - month
                asset_b:
                    batch_identifiers:
                        - day
                        - month
                        - year
        """,
        ),
        runtime_environment={"name": "my_datasource"},
        config_defaults={"module_name": "great_expectations.datasource"},
    )
    return basic_datasource


####################################
# Tests with data passed in as query
####################################


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_self_check(
    db_file, datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    report = (
        datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.self_check()
    )

    assert report == {
        "execution_engine": {
            "connection_string": f"sqlite:///{db_file}",
            "module_name": "great_expectations.execution_engine.sqlalchemy_execution_engine",
            "class_name": "SqlAlchemyExecutionEngine",
        },
        "data_connectors": {
            "count": 1,
            "test_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "data_asset_count": 2,
                "data_assets": {
                    "asset_a": {
                        "batch_definition_count": 0,
                        "example_data_references": [],
                    },
                    "asset_b": {
                        "batch_definition_count": 0,
                        "example_data_references": [],
                    },
                },
                "example_data_asset_names": ["asset_a", "asset_b"],
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_available_data_asset(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    expected_available_data_asset_names: Dict[List[str]] = {
        "test_runtime_data_connector": ["asset_a", "asset_b"]
    }
    available_data_asset_names: Dict[
        List[str]
    ] = (
        datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_available_data_asset_names()
    )
    assert available_data_asset_names == expected_available_data_asset_names


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_unknown_datasource(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"
    # Test for an unknown datasource
    with pytest.raises(TypeError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=RuntimeBatchRequest(
                datasource_name="non_existent_datasource",
                data_connector_name="test_runtime_data_connector",
                data_asset_name="my_data_asset",
                runtime_parameters={"query": test_query},
                batch_identifiers=None,
            )
        )


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_unknown_dataconnector(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"
    # Test for an unknown data_connector
    with pytest.raises(TypeError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=RuntimeBatchRequest(
                datasource_name=datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
                data_connector_name="non_existent_data_connector",
                data_asset_name="my_data_asset",
                runtime_parameters={"query": test_query},
                batch_identifiers=None,
            )
        )


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_no_batch_identifiers(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    # Test for illegal absence of batch_identifiers when batch_data is specified
    with pytest.raises(TypeError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=RuntimeBatchRequest(
                datasource_name=datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
                data_connector_name="test_runtime_data_connector",
                data_asset_name="my_data_asset",
                runtime_parameters={"query": test_query},
                batch_identifiers=None,
            )
        )


def test_datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine_illegal_batch_identifiers(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    # Test for illegal falsiness of batch_identifiers when batch_data is specified
    with pytest.raises(TypeError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=RuntimeBatchRequest(
                datasource_name=datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
                data_connector_name="test_runtime_data_connector",
                data_asset_name="my_data_asset",
                runtime_parameters={"query": test_query},
                batch_identifiers={},
            )
        )


def test_batch_identifiers_and_batch_identifiers_success_all_keys_present_with_query(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"
    batch_identifiers: Dict[str, Union[str, int]] = {
        "pipeline_stage_name": "core_processing",
        "airflow_run_id": 1234567890,
        "custom_key_0": "custom_value_0",
    }

    # Verify that all keys in batch_identifiers are acceptable as batch_identifiers (using batch count).
    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "TEMP_QUERY_DATA_ASSET",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    assert len(batch_list[0].head().columns) == 11


def test_batch_identifiers_and_batch_identifiers_success_no_temp_table(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I"
    batch_identifiers: Dict[str, Union[str, int]] = {
        "pipeline_stage_name": "core_processing",
        "airflow_run_id": 1234567890,
        "custom_key_0": "custom_value_0",
    }

    # Verify that all keys in batch_identifiers are acceptable as batch_identifiers (using batch count).
    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "TEMP_QUERY_DATA_ASSET",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
        "batch_spec_passthrough": {"create_temp_table": False},
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    assert len(batch_list[0].head().columns) == 11


def test_batch_identifiers_and_batch_identifiers_error_illegal_key_with_query_mostly_legal_keys(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I"
    batch_identifiers: Dict[str, Union[str, int]] = {
        "pipeline_stage_name": "core_processing",
        "airflow_run_id": 1234567890,
        "custom_key_0": "custom_value_0",
        "i_am_illegal_key": "i_am_illegal_value",
    }

    # Insure that keys in batch_identifiers that are not among batch_identifiers declared in
    # configuration are not accepted.  In this test, all legal keys plus a single illegal key are present.
    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "TEMP_QUERY_DATA_ASSET",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)

    with pytest.raises(gx_exceptions.DataConnectorError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=batch_request
        )


def test_batch_identifiers_and_batch_identifiers_error_illegal_key_with_query_single_illegal_key(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"
    batch_identifiers: Dict[str, Union[str, int]] = {"unknown_key": "some_value"}
    # Ensure that keys in batch_identifiers that are not among batch_identifiers declared in
    # configuration  are not accepted.  In this test, a single illegal key is present.
    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "TEMP_QUERY_DATA_ASSET",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)

    with pytest.raises(gx_exceptions.DataConnectorError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=batch_request
        )


def test_set_data_asset_name_for_runtime_query_data(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    test_query: str = "SELECT * FROM table_full__I;"
    batch_identifiers: Dict[str, Union[str, int]] = {
        "pipeline_stage_name": "core_processing",
        "airflow_run_id": 1234567890,
        "custom_key_0": "custom_value_0",
    }

    # set : my_runtime_data_asset
    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "my_runtime_data_asset",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    assert batch_list[0].batch_definition.data_asset_name == "my_runtime_data_asset"


def test_get_batch_definition_list_from_batch_request_length_one_from_query(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    batch_identifiers: Dict[str, Union[str, int]] = {
        "airflow_run_id": 1234567890,
    }

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "my_data_asset",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    # batches are a little bit more difficult to test because of batch_markers
    # they are ones that uniquely identify the data
    assert len(batch_list) == 1
    my_batch_1 = batch_list[0]
    assert my_batch_1.batch_spec is not None
    assert my_batch_1.batch_definition["data_asset_name"] == "my_data_asset"
    assert isinstance(my_batch_1.data.selectable, sqlalchemy.Table)


def test_get_batch_list_from_batch_request_length_one_from_query_named_asset(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    batch_identifiers: Dict[str, Union[str, int]] = {"day": 2, "month": 12}

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "asset_a",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    # batches are a little bit more difficult to test because of batch_markers
    # they are ones that uniquely identify the data
    assert len(batch_list) == 1
    my_batch_1 = batch_list[0]
    assert my_batch_1.batch_spec is not None
    assert my_batch_1.batch_definition["data_asset_name"] == "asset_a"
    assert isinstance(my_batch_1.data.selectable, sqlalchemy.Table)


def test_get_batch_list_from_batch_request_length_one_from_query_named_asset_two_batch_requests(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    batch_identifiers: Dict[str, Union[str, int]] = {"day": 1, "month": 12}

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "asset_a",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    # batches are a little bit more difficult to test because of batch_markers
    # they are ones that uniquely identify the data
    assert len(batch_list) == 1
    my_batch_1 = batch_list[0]
    assert my_batch_1.batch_spec is not None
    assert my_batch_1.batch_definition["data_asset_name"] == "asset_a"
    assert isinstance(my_batch_1.data.selectable, sqlalchemy.Table)

    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    batch_identifiers: Dict[str, Union[str, int]] = {"day": 2, "month": 12}

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": "test_runtime_data_connector",
        "data_asset_name": "asset_a",
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": batch_identifiers,
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)
    batch_list: List[
        Batch
    ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
        batch_request=batch_request
    )
    # batches are a little bit more difficult to test because of batch_markers
    # they are ones that uniquely identify the data
    assert len(batch_list) == 1
    my_batch_2 = batch_list[0]
    assert my_batch_2.batch_spec is not None
    assert my_batch_2.batch_definition["data_asset_name"] == "asset_a"
    assert isinstance(my_batch_2.data.selectable, sqlalchemy.Table)


def test_get_batch_with_pipeline_style_batch_request_missing_batch_identifiers_error(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    data_connector_name: str = "test_runtime_data_connector"
    data_asset_name: str = "test_asset_1"

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": data_connector_name,
        "data_asset_name": data_asset_name,
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": None,
    }

    with pytest.raises(TypeError):
        # noinspection PyUnusedLocal
        batch_list: List[
            Batch
        ] = datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_batch_list_from_batch_request(
            batch_request=RuntimeBatchRequest(**batch_request)
        )


def test_get_batch_definitions_and_get_batch_basics_from_query(
    datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine, sa
):
    # interacting with the database using query
    test_query: str = "SELECT * FROM table_full__I;"

    data_connector_name: str = "test_runtime_data_connector"
    data_asset_name: str = "test_asset_1"

    batch_request: Dict[str, Any] = {
        "datasource_name": datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.name,
        "data_connector_name": data_connector_name,
        "data_asset_name": data_asset_name,
        "runtime_parameters": {
            "query": test_query,
        },
        "batch_identifiers": {
            "airflow_run_id": 1234567890,
        },
    }
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(**batch_request)

    assert (
        len(
            datasource_with_runtime_data_connector_and_sqlalchemy_execution_engine.get_available_batch_definitions(
                batch_request=batch_request
            )
        )
        == 1
    )



import datetime
import itertools
import json
import os
import tempfile
from unittest import mock

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations import DataContext
from great_expectations.core import ExpectationSuite
from great_expectations.data_context.config_validator.yaml_config_validator import (
    _YamlConfigValidator,
)
from great_expectations.data_context.store import CheckpointStore
from great_expectations.data_context.util import file_relative_path
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler
from great_expectations.util import get_sqlalchemy_url, load_class
from tests.core.usage_statistics.util import (
    usage_stats_exceptions_exist,
    usage_stats_invalid_messages_exist,
)
from tests.test_utils import create_files_in_directory, set_directory


@pytest.fixture
def test_connectable_postgresql_db(sa, test_backends, test_df):
    """Populates a postgres DB with a `test_df` table in the `connection_test` schema to test DataConnectors against"""

    if "postgresql" not in test_backends:
        pytest.skip("skipping fixture because postgresql not selected")

    url = get_sqlalchemy_url(
        drivername="postgresql",
        username="postgres",
        password="",
        host=os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        port="5432",
        database="test_ci",
    )
    engine = sa.create_engine(url)

    schema_check_results = engine.execute(
        "SELECT schema_name FROM information_schema.schemata WHERE schema_name = 'connection_test';"
    ).fetchall()
    if len(schema_check_results) == 0:
        engine.execute("CREATE SCHEMA connection_test;")

    table_check_results = engine.execute(
        """
SELECT EXISTS (
   SELECT FROM information_schema.tables
   WHERE  table_schema = 'connection_test'
   AND    table_name   = 'test_df'
);
"""
    ).fetchall()
    if table_check_results != [(True,)]:
        test_df.to_sql(name="test_df", con=engine, index=True, schema="connection_test")

    # Return a connection string to this newly-created db
    return engine


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_config_with_yaml_error(mock_emit, caplog, empty_data_context_stats_enabled):
    with pytest.raises(Exception):
        # noinspection PyUnusedLocal
        my_expectation_store = empty_data_context_stats_enabled.test_yaml_config(
            yaml_config="""
module_name: great_expectations.data_context.store.expectations_store
class_name: ExpectationsStore
store_backend:
    module_name: "great_expectations.data_context.store.store_backend"
    class_name: InMemoryStoreBackend
EGREGIOUS FORMATTING ERROR
"""
        )
    assert mock_emit.call_count == 1
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {"diagnostic_info": ["__yaml_parse_error__"]},
                "success": False,
            }
        ),
    ]

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.slow  # 1.61s
def test_expectations_store_with_filesystem_store_backend(
    mock_emit, caplog, empty_data_context_stats_enabled
):
    tmp_dir = str(tempfile.mkdtemp())
    with open(os.path.join(tmp_dir, "expectations_A1.json"), "w") as f_:
        f_.write("\n")
    with open(os.path.join(tmp_dir, "expectations_A2.json"), "w") as f_:
        f_.write("\n")

    # noinspection PyUnusedLocal
    my_expectation_store = empty_data_context_stats_enabled.test_yaml_config(
        yaml_config=f"""
module_name: great_expectations.data_context.store
class_name: ExpectationsStore
store_backend:
    module_name: "great_expectations.data_context.store"
    class_name: TupleFilesystemStoreBackend
    base_directory: {tmp_dir}
"""
    )
    assert mock_emit.call_count == 1
    # Substitute current anonymized name since it changes for each run
    anonymized_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_name"
    ]
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "ExpectationsStore",
                    "anonymized_store_backend": {
                        "parent_class": "TupleFilesystemStoreBackend"
                    },
                },
                "success": True,
            }
        )
    ]

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_checkpoint_store_with_filesystem_store_backend(
    mock_emit, caplog, empty_data_context_stats_enabled, tmp_path_factory
):
    tmp_dir: str = str(
        tmp_path_factory.mktemp("test_checkpoint_store_with_filesystem_store_backend")
    )
    context: DataContext = empty_data_context_stats_enabled

    yaml_config: str = f"""
    store_name: my_checkpoint_store
    class_name: CheckpointStore
    module_name: great_expectations.data_context.store
    store_backend:
        class_name: TupleFilesystemStoreBackend
        module_name: "great_expectations.data_context.store"
        base_directory: {tmp_dir}/checkpoints
    """

    my_checkpoint_store: CheckpointStore = context.test_yaml_config(
        yaml_config=yaml_config,
        return_mode="instantiated_class",
    )
    assert mock_emit.call_count == 1
    # Substitute anonymized_name since it changes for each run
    anonymized_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_name"
    ]
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "CheckpointStore",
                    "anonymized_store_backend": {
                        "parent_class": "TupleFilesystemStoreBackend"
                    },
                },
                "success": True,
            }
        ),
    ]

    report_object: dict = context.test_yaml_config(
        yaml_config=yaml_config,
        return_mode="report_object",
    )
    assert mock_emit.call_count == 2
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "CheckpointStore",
                    "anonymized_store_backend": {
                        "parent_class": "TupleFilesystemStoreBackend"
                    },
                },
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "CheckpointStore",
                    "anonymized_store_backend": {
                        "parent_class": "TupleFilesystemStoreBackend"
                    },
                },
                "success": True,
            }
        ),
    ]

    assert my_checkpoint_store.config == report_object["config"]

    expected_checkpoint_store_config: dict

    expected_checkpoint_store_config = {
        "store_name": "my_checkpoint_store",
        "class_name": "CheckpointStore",
        "module_name": "great_expectations.data_context.store.checkpoint_store",
        "store_backend": {
            "module_name": "great_expectations.data_context.store",
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": f"{tmp_dir}/checkpoints",
            "suppress_store_backend_id": True,
            "filepath_suffix": ".yml",
        },
        "overwrite_existing": False,
        "runtime_environment": {
            "root_directory": f"{context.root_directory}",
        },
    }
    assert my_checkpoint_store.config == expected_checkpoint_store_config

    checkpoint_store_name: str = my_checkpoint_store.config["store_name"]
    context.get_config()["checkpoint_store_name"] = checkpoint_store_name

    assert (
        context.get_config_with_variables_substituted().checkpoint_store_name
        == "my_checkpoint_store"
    )
    assert (
        context.get_config_with_variables_substituted().checkpoint_store_name
        == my_checkpoint_store.config["store_name"]
    )

    expected_checkpoint_store_config = {
        "store_name": "my_checkpoint_store",
        "class_name": "CheckpointStore",
        "module_name": "great_expectations.data_context.store",
        "store_backend": {
            "class_name": "TupleFilesystemStoreBackend",
            "module_name": "great_expectations.data_context.store",
            "base_directory": f"{tmp_dir}/checkpoints",
            "suppress_store_backend_id": True,
        },
    }
    assert (
        context.get_config_with_variables_substituted().stores[
            context.get_config_with_variables_substituted().checkpoint_store_name
        ]
        == expected_checkpoint_store_config
    )
    # No other usage stats calls
    assert mock_emit.call_count == 2

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@pytest.mark.filterwarnings(
    "ignore:String run_ids are deprecated*:DeprecationWarning:great_expectations.data_context.types.resource_identifiers"
)
@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_empty_store2(mock_emit, caplog, empty_data_context_stats_enabled):
    empty_data_context_stats_enabled.test_yaml_config(
        yaml_config="""
class_name: ValidationsStore
store_backend:

    module_name: "great_expectations.data_context.store.store_backend"
    class_name: InMemoryStoreBackend
"""
    )
    assert mock_emit.call_count == 1
    # Substitute anonymized_name since it changes for each run
    anonymized_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_name"
    ]
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "ValidationsStore",
                    "anonymized_store_backend": {
                        "parent_class": "InMemoryStoreBackend"
                    },
                },
                "success": True,
            }
        ),
    ]

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_datasource_config(mock_emit, caplog, empty_data_context_stats_enabled):
    temp_dir = str(tempfile.mkdtemp())
    create_files_in_directory(
        directory=temp_dir,
        file_name_list=[
            "alex_20200809_1000.csv",
            "eugene_20200809_1500.csv",
            "james_20200811_1009.csv",
            "abe_20200809_1040.csv",
            "will_20200809_1002.csv",
            "james_20200713_1567.csv",
            "eugene_20201129_1900.csv",
            "will_20200810_1001.csv",
            "james_20200810_1003.csv",
            "alex_20200819_1300.csv",
        ],
    )
    print(temp_dir)

    return_obj = empty_data_context_stats_enabled.test_yaml_config(
        yaml_config=f"""
class_name: Datasource

execution_engine:
    class_name: PandasExecutionEngine

data_connectors:
    my_filesystem_data_connector:
        # class_name: ConfiguredAssetFilesystemDataConnector
        class_name: InferredAssetFilesystemDataConnector
        base_directory: {temp_dir}
        glob_directive: '*.csv'
        default_regex:
            pattern: (.+)_(\\d+)\\.csv
            group_names:
            - letter
            - number
""",
        return_mode="report_object",
    )

    # Test usage stats messages
    assert mock_emit.call_count == 1
    # Substitute anonymized names since it changes for each run
    anonymized_datasource_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_name"
    ]
    anonymized_execution_engine_name = mock_emit.call_args_list[0][0][0][
        "event_payload"
    ]["anonymized_execution_engine"]["anonymized_name"]
    anonymized_data_connector_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_data_connectors"
    ][0]["anonymized_name"]
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_datasource_name,
                    "parent_class": "Datasource",
                    "anonymized_execution_engine": {
                        "anonymized_name": anonymized_execution_engine_name,
                        "parent_class": "PandasExecutionEngine",
                    },
                    "anonymized_data_connectors": [
                        {
                            "anonymized_name": anonymized_data_connector_name,
                            "parent_class": "InferredAssetFilesystemDataConnector",
                        }
                    ],
                },
                "success": True,
            }
        )
    ]

    print(json.dumps(return_obj, indent=2))

    assert set(return_obj.keys()) == {"execution_engine", "data_connectors"}
    sub_obj = return_obj["data_connectors"]["my_filesystem_data_connector"]
    # FIXME: (Sam) example_data_reference removed temporarily in PR #2590:
    # sub_obj.pop("example_data_reference")
    # assert sub_obj == {
    #     "class_name": "InferredAssetFilesystemDataConnector",
    #     "data_asset_count": 1,
    #     "example_data_asset_names": ["DEFAULT_ASSET_NAME"],
    #     "data_assets": {
    #         "DEFAULT_ASSET_NAME": {
    #             "batch_definition_count": 10,
    #             "example_data_references": [
    #                 "abe_20200809_1040.csv",
    #                 "alex_20200809_1000.csv",
    #                 "alex_20200819_1300.csv",
    #             ],
    #         }
    #     },
    #     "example_unmatched_data_references": [],
    #     "unmatched_data_reference_count": 0,
    # }
    # No other usage stats calls
    assert mock_emit.call_count == 1

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_error_states(mock_emit, caplog, empty_data_context_stats_enabled):
    first_config: str = """
class_name: Datasource

execution_engine:
    class_name: NOT_A_REAL_CLASS_NAME
"""

    with pytest.raises(gx_exceptions.DatasourceInitializationError) as excinfo:
        empty_data_context_stats_enabled.test_yaml_config(yaml_config=first_config)
        # print(excinfo.value.message)
        # shortened_message_len = len(excinfo.value.message)
        # print("="*80)

    assert mock_emit.call_count == 1
    expected_call_args_list = [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {"parent_class": "Datasource"},
                "success": False,
            }
        ),
    ]
    assert mock_emit.call_args_list == expected_call_args_list

    # Set shorten_tracebacks=True and verify that no error is thrown, even though the config is the same as before.
    # Note: a more thorough test could also verify that the traceback is indeed short.
    empty_data_context_stats_enabled.test_yaml_config(
        yaml_config=first_config,
        shorten_tracebacks=True,
    )
    assert mock_emit.call_count == 2
    expected_call_args_list.append(
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {"parent_class": "Datasource"},
                "success": False,
            }
        ),
    )
    assert mock_emit.call_args_list == expected_call_args_list

    # For good measure, do it again, with a different config and a different type of error
    # Note this erroneous key/value does not cause an error and is removed from the Datasource config
    temp_dir = str(tempfile.mkdtemp())
    second_config = f"""
class_name: Datasource

execution_engine:
    class_name: PandasExecutionEngine

data_connectors:
    my_filesystem_data_connector:
        # class_name: ConfiguredAssetFilesystemDataConnector
        class_name: InferredAssetFilesystemDataConnector
        base_directory: {temp_dir}
        glob_directive: '*.csv'
        default_regex:
            pattern: (.+)_(\\d+)\\.csv
            group_names:
            - letter
            - number
        NOT_A_REAL_KEY: nothing
"""

    datasource = empty_data_context_stats_enabled.test_yaml_config(
        yaml_config=second_config
    )
    assert (
        "NOT_A_REAL_KEY"
        not in datasource.config["data_connectors"]["my_filesystem_data_connector"]
    )
    assert mock_emit.call_count == 3
    # Substitute anonymized names since it changes for each run
    anonymized_datasource_name = mock_emit.call_args_list[2][0][0]["event_payload"][
        "anonymized_name"
    ]
    anonymized_execution_engine_name = mock_emit.call_args_list[2][0][0][
        "event_payload"
    ]["anonymized_execution_engine"]["anonymized_name"]
    anonymized_data_connector_name = mock_emit.call_args_list[2][0][0]["event_payload"][
        "anonymized_data_connectors"
    ][0]["anonymized_name"]
    expected_call_args_list.append(
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_datasource_name,
                    "parent_class": "Datasource",
                    "anonymized_execution_engine": {
                        "anonymized_name": anonymized_execution_engine_name,
                        "parent_class": "PandasExecutionEngine",
                    },
                    "anonymized_data_connectors": [
                        {
                            "anonymized_name": anonymized_data_connector_name,
                            "parent_class": "InferredAssetFilesystemDataConnector",
                        }
                    ],
                },
                "success": True,
            }
        ),
    )
    assert mock_emit.call_args_list == expected_call_args_list

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_config_variables_in_test_yaml_config(
    mock_emit, caplog, empty_data_context_stats_enabled, sa
):
    context: DataContext = empty_data_context_stats_enabled

    db_file = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )

    context.save_config_variable("db_file", db_file)
    context.save_config_variable(
        "data_connector_name", "my_very_awesome_data_connector"
    )
    context.save_config_variable("suffix", "__whole_table")
    context.save_config_variable("sampling_n", "10")

    print(context.config_variables)

    first_config = """
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///${db_file}

introspection:
    ${data_connector_name}:
        data_asset_name_suffix: ${suffix}
        sampling_method: _sample_using_limit
        sampling_kwargs:
            n: ${sampling_n}
"""

    my_datasource = context.test_yaml_config(first_config)
    assert (
        "test_cases_for_sql_data_connector.db"
        in my_datasource.execution_engine.connection_string
    )
    assert mock_emit.call_count == 1
    # Substitute anonymized names since it changes for each run
    anonymized_datasource_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_name"
    ]
    anonymized_data_connector_name = mock_emit.call_args_list[0][0][0]["event_payload"][
        "anonymized_data_connectors"
    ][0]["anonymized_name"]
    expected_call_args_list = [
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_datasource_name,
                    "parent_class": "SimpleSqlalchemyDatasource",
                    "anonymized_execution_engine": {
                        "parent_class": "SqlAlchemyExecutionEngine"
                    },
                    "anonymized_data_connectors": [
                        {
                            "anonymized_name": anonymized_data_connector_name,
                            "parent_class": "InferredAssetSqlDataConnector",
                        }
                    ],
                },
                "success": True,
            }
        ),
    ]
    assert mock_emit.call_args_list == expected_call_args_list

    report_object = context.test_yaml_config(first_config, return_mode="report_object")
    print(json.dumps(report_object, indent=2))
    assert report_object["data_connectors"]["count"] == 1
    assert set(report_object["data_connectors"].keys()) == {
        "count",
        "${data_connector_name}",
    }
    assert mock_emit.call_count == 2
    expected_call_args_list.append(
        mock.call(
            {
                "event": "data_context.test_yaml_config",
                "event_payload": {
                    "anonymized_name": anonymized_datasource_name,
                    "parent_class": "SimpleSqlalchemyDatasource",
                    "anonymized_execution_engine": {
                        "parent_class": "SqlAlchemyExecutionEngine"
                    },
                    "anonymized_data_connectors": [
                        {
                            "anonymized_name": anonymized_data_connector_name,
                            "parent_class": "InferredAssetSqlDataConnector",
                        }
                    ],
                },
                "success": True,
            }
        ),
    )
    assert mock_emit.call_args_list == expected_call_args_list

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_golden_path_sql_datasource_configuration(
    mock_emit,
    caplog,
    empty_data_context_stats_enabled,
    sa,
    test_connectable_postgresql_db,
):
    """Tests the golden path for setting up a StreamlinedSQLDatasource using test_yaml_config"""
    context: DataContext = empty_data_context_stats_enabled

    with set_directory(context.root_directory):

        # Everything below this line (except for asserts) is what we expect users to run as part of the golden path.
        import great_expectations as gx

        context = gx.get_context()

        db_hostname = os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost")
        yaml_config = f"""
    class_name: SimpleSqlalchemyDatasource
    credentials:
        drivername: postgresql
        username: postgres
        password: ""
        host: {db_hostname}
        port: 5432
        database: test_ci

    introspection:
        whole_table_with_limits:
            sampling_method: _sample_using_limit
            sampling_kwargs:
                n: 10
    """
        # noinspection PyUnusedLocal
        report_object = context.test_yaml_config(
            name="my_datasource",
            yaml_config=yaml_config,
            return_mode="report_object",
        )
        assert mock_emit.call_count == 2
        # Substitute anonymized names since it changes for each run
        anonymized_datasource_name = mock_emit.call_args_list[1][0][0]["event_payload"][
            "anonymized_name"
        ]
        anonymized_data_connector_name = mock_emit.call_args_list[1][0][0][
            "event_payload"
        ]["anonymized_data_connectors"][0]["anonymized_name"]
        expected_call_args_list = [
            mock.call(
                {"event_payload": {}, "event": "data_context.__init__", "success": True}
            ),
            mock.call(
                {
                    "event": "data_context.test_yaml_config",
                    "event_payload": {
                        "anonymized_name": anonymized_datasource_name,
                        "parent_class": "SimpleSqlalchemyDatasource",
                        "anonymized_execution_engine": {
                            "parent_class": "SqlAlchemyExecutionEngine"
                        },
                        "anonymized_data_connectors": [
                            {
                                "anonymized_name": anonymized_data_connector_name,
                                "parent_class": "InferredAssetSqlDataConnector",
                            }
                        ],
                    },
                    "success": True,
                }
            ),
        ]
        assert mock_emit.call_args_list == expected_call_args_list

        print(json.dumps(report_object, indent=2))
        print(context.datasources)

        my_batch = context.get_batch(
            "my_datasource",
            "whole_table_with_limits",
            "test_df",
        )
        # assert len(my_batch.data.fetchall()) == 10

        with pytest.raises(KeyError):
            my_batch = context.get_batch(
                "my_datasource",
                "whole_table_with_limits",
                "DOES_NOT_EXIST",
            )

        my_validator = context.get_validator(
            datasource_name="my_datasource",
            data_connector_name="whole_table_with_limits",
            data_asset_name="test_df",
            expectation_suite=ExpectationSuite(
                "my_expectation_suite", data_context=context
            ),
        )
        my_evr = my_validator.expect_table_columns_to_match_set(column_set=[])
        print(my_evr)

        # my_evr = my_validator.expect_column_values_to_be_between(
        #     column="x",
        #     min_value=0,
        #     max_value=4,
        # )
        # assert my_evr.success

        # TODO: <Alex>ALEX</Alex>
        # my_evr = my_validator.expect_table_columns_to_match_ordered_list(ordered_list=["a", "b", "c"])
        # assert my_evr.success

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.filterwarnings(
    "ignore:get_batch is deprecated*:DeprecationWarning:great_expectations.data_context.data_context"
)
def test_golden_path_inferred_asset_pandas_datasource_configuration(
    mock_emit, caplog, empty_data_context_stats_enabled, test_df, tmp_path_factory
):
    """
    Tests the golden path for InferredAssetFilesystemDataConnector with PandasExecutionEngine using test_yaml_config
    """
    base_directory = str(
        tmp_path_factory.mktemp("test_golden_path_pandas_datasource_configuration")
    )

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_charlie/A/A-1.csv",
            "test_dir_charlie/A/A-2.csv",
            "test_dir_charlie/A/A-3.csv",
            "test_dir_charlie/B/B-1.csv",
            "test_dir_charlie/B/B-2.csv",
            "test_dir_charlie/B/B-3.csv",
            "test_dir_charlie/C/C-1.csv",
            "test_dir_charlie/C/C-2.csv",
            "test_dir_charlie/C/C-3.csv",
            "test_dir_charlie/D/D-1.csv",
            "test_dir_charlie/D/D-2.csv",
            "test_dir_charlie/D/D-3.csv",
        ],
        file_content_fn=lambda: test_df.to_csv(header=True, index=False),
    )

    context: DataContext = empty_data_context_stats_enabled

    with set_directory(context.root_directory):
        import great_expectations as gx

        context = gx.get_context()
        mock_emit.reset_mock()  # Remove data_context.__init__ call

        yaml_config = f"""
    class_name: Datasource

    execution_engine:
        class_name: PandasExecutionEngine

    data_connectors:
        my_filesystem_data_connector:
            class_name: InferredAssetFilesystemDataConnector
            base_directory: {base_directory}/test_dir_charlie
            glob_directive: "*/*.csv"

            default_regex:
                pattern: (.+)/(.+)-(\\d+)\\.csv
                group_names:
                    - subdirectory
                    - data_asset_name
                    - number
    """

        # noinspection PyUnusedLocal
        report_object = context.test_yaml_config(
            name="my_directory_datasource",
            yaml_config=yaml_config,
            return_mode="report_object",
        )
        # print(json.dumps(report_object, indent=2))
        # print(context.datasources)
        assert mock_emit.call_count == 1
        # Substitute anonymized names since it changes for each run
        anonymized_datasource_name = mock_emit.call_args_list[0][0][0]["event_payload"][
            "anonymized_name"
        ]
        anonymized_execution_engine_name = mock_emit.call_args_list[0][0][0][
            "event_payload"
        ]["anonymized_execution_engine"]["anonymized_name"]
        anonymized_data_connector_name = mock_emit.call_args_list[0][0][0][
            "event_payload"
        ]["anonymized_data_connectors"][0]["anonymized_name"]
        expected_call_args_list = [
            mock.call(
                {
                    "event": "data_context.test_yaml_config",
                    "event_payload": {
                        "anonymized_name": anonymized_datasource_name,
                        "parent_class": "Datasource",
                        "anonymized_execution_engine": {
                            "anonymized_name": anonymized_execution_engine_name,
                            "parent_class": "PandasExecutionEngine",
                        },
                        "anonymized_data_connectors": [
                            {
                                "anonymized_name": anonymized_data_connector_name,
                                "parent_class": "InferredAssetFilesystemDataConnector",
                            }
                        ],
                    },
                    "success": True,
                }
            ),
        ]
        assert mock_emit.call_args_list == expected_call_args_list

        my_batch = context.get_batch(
            datasource_name="my_directory_datasource",
            data_connector_name="my_filesystem_data_connector",
            data_asset_name="A",
            batch_identifiers={
                "number": "2",
            },
            batch_spec_passthrough={
                "sampling_method": "_sample_using_hash",
                "sampling_kwargs": {
                    "column_name": "date",
                    "hash_function_name": "md5",
                    "hash_value": "f",
                },
            },
        )
        assert my_batch.batch_definition["data_asset_name"] == "A"

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 2

        df_data = my_batch.data.dataframe
        assert df_data.shape == (10, 10)
        df_data["date"] = df_data.apply(
            lambda row: datetime.datetime.strptime(row["date"], "%Y-%m-%d").date(),
            axis=1,
        )
        assert (
            test_df[
                (test_df["date"] == datetime.date(2020, 1, 15))
                | (test_df["date"] == datetime.date(2020, 1, 29))
            ]
            .drop("timestamp", axis=1)
            .equals(df_data.drop("timestamp", axis=1))
        )

        with pytest.raises(ValueError):
            # noinspection PyUnusedLocal
            my_batch = context.get_batch(
                datasource_name="my_directory_datasource",
                data_connector_name="my_filesystem_data_connector",
                data_asset_name="DOES_NOT_EXIST",
            )

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 3

        my_validator = context.get_validator(
            datasource_name="my_directory_datasource",
            data_connector_name="my_filesystem_data_connector",
            data_asset_name="D",
            data_connector_query={"batch_filter_parameters": {"number": "3"}},
            expectation_suite=ExpectationSuite(
                "my_expectation_suite", data_context=context
            ),
            batch_spec_passthrough={
                "sampling_method": "_sample_using_hash",
                "sampling_kwargs": {
                    "column_name": "date",
                    "hash_function_name": "md5",
                    "hash_value": "f",
                },
            },
        )

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 4

        my_evr = my_validator.expect_column_values_to_be_between(
            column="d", min_value=1, max_value=31
        )
        assert my_evr.success

        # TODO: <Alex>ALEX</Alex>
        # my_evr = my_validator.expect_table_columns_to_match_ordered_list(ordered_list=["x", "y", "z"])
        # assert my_evr.success

        # No other usage stats calls detected
        # assert mock_emit.call_count == 1
        assert mock_emit.call_count == 4

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.filterwarnings(
    "ignore:get_batch is deprecated*:DeprecationWarning:great_expectations.data_context.data_context"
)
def test_golden_path_configured_asset_pandas_datasource_configuration(
    mock_emit, caplog, empty_data_context_stats_enabled, test_df, tmp_path_factory
):
    """
    Tests the golden path for InferredAssetFilesystemDataConnector with PandasExecutionEngine using test_yaml_config
    """
    base_directory = str(
        tmp_path_factory.mktemp("test_golden_path_pandas_datasource_configuration")
    )

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_foxtrot/A/A-1.csv",
            "test_dir_foxtrot/A/A-2.csv",
            "test_dir_foxtrot/A/A-3.csv",
            "test_dir_foxtrot/B/B-1.txt",
            "test_dir_foxtrot/B/B-2.txt",
            "test_dir_foxtrot/B/B-3.txt",
            "test_dir_foxtrot/C/C-2017.csv",
            "test_dir_foxtrot/C/C-2018.csv",
            "test_dir_foxtrot/C/C-2019.csv",
            "test_dir_foxtrot/D/D-aaa.csv",
            "test_dir_foxtrot/D/D-bbb.csv",
            "test_dir_foxtrot/D/D-ccc.csv",
            "test_dir_foxtrot/D/D-ddd.csv",
            "test_dir_foxtrot/D/D-eee.csv",
        ],
        file_content_fn=lambda: test_df.to_csv(header=True, index=False),
    )

    context: DataContext = empty_data_context_stats_enabled

    with set_directory(context.root_directory):
        import great_expectations as gx

        context = gx.get_context()
        mock_emit.reset_mock()  # Remove data_context.__init__ call

        yaml_config = f"""
    class_name: Datasource

    execution_engine:
        class_name: PandasExecutionEngine

    data_connectors:
        my_filesystem_data_connector:
            class_name: ConfiguredAssetFilesystemDataConnector
            base_directory: {base_directory}
            # glob_directive: "*"

            default_regex:
                pattern: (.+)\\.csv
                group_names:
                    - alphanumeric

            assets:
                A:
                    base_directory: {base_directory}/test_dir_foxtrot/A
                    pattern: (.+)-(\\d+)\\.csv
                    group_names:
                        - letter
                        - number
                B:
                    base_directory: {base_directory}/test_dir_foxtrot/B
                    pattern: (.+)-(\\d+)\\.csv
                    group_names:
                        - letter
                        - number
                C:
                    base_directory: {base_directory}/test_dir_foxtrot/C
                    pattern: (.+)-(\\d+)\\.csv
                    group_names:
                        - letter
                        - year
                D:
                    base_directory: {base_directory}/test_dir_foxtrot/D
                    pattern: (.+)-(\\d+)\\.csv
                    group_names:
                        - letter
                        - checksum
    """

        # noinspection PyUnusedLocal
        report_object = context.test_yaml_config(
            name="my_directory_datasource",
            yaml_config=yaml_config,
            return_mode="report_object",
        )
        # print(json.dumps(report_object, indent=2))
        # print(context.datasources)
        assert mock_emit.call_count == 1
        # Substitute anonymized names since it changes for each run
        anonymized_datasource_name = mock_emit.call_args_list[0][0][0]["event_payload"][
            "anonymized_name"
        ]
        anonymized_execution_engine_name = mock_emit.call_args_list[0][0][0][
            "event_payload"
        ]["anonymized_execution_engine"]["anonymized_name"]
        anonymized_data_connector_name = mock_emit.call_args_list[0][0][0][
            "event_payload"
        ]["anonymized_data_connectors"][0]["anonymized_name"]
        expected_call_args_list = [
            mock.call(
                {
                    "event": "data_context.test_yaml_config",
                    "event_payload": {
                        "anonymized_name": anonymized_datasource_name,
                        "parent_class": "Datasource",
                        "anonymized_execution_engine": {
                            "anonymized_name": anonymized_execution_engine_name,
                            "parent_class": "PandasExecutionEngine",
                        },
                        "anonymized_data_connectors": [
                            {
                                "anonymized_name": anonymized_data_connector_name,
                                "parent_class": "ConfiguredAssetFilesystemDataConnector",
                            }
                        ],
                    },
                    "success": True,
                }
            ),
        ]
        assert mock_emit.call_args_list == expected_call_args_list

        my_batch = context.get_batch(
            datasource_name="my_directory_datasource",
            data_connector_name="my_filesystem_data_connector",
            data_asset_name="A",
            batch_identifiers={
                "number": "2",
            },
            batch_spec_passthrough={
                "sampling_method": "_sample_using_hash",
                "sampling_kwargs": {
                    "column_name": "date",
                    "hash_function_name": "md5",
                    "hash_value": "f",
                },
            },
        )
        assert my_batch.batch_definition["data_asset_name"] == "A"

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 2

        my_batch.head()

        df_data = my_batch.data.dataframe
        assert df_data.shape == (10, 10)
        df_data["date"] = df_data.apply(
            lambda row: datetime.datetime.strptime(row["date"], "%Y-%m-%d").date(),
            axis=1,
        )
        assert (
            test_df[
                (test_df["date"] == datetime.date(2020, 1, 15))
                | (test_df["date"] == datetime.date(2020, 1, 29))
            ]
            .drop("timestamp", axis=1)
            .equals(df_data.drop("timestamp", axis=1))
        )

        with pytest.raises(ValueError):
            # noinspection PyUnusedLocal
            my_batch = context.get_batch(
                datasource_name="my_directory_datasource",
                data_connector_name="my_filesystem_data_connector",
                data_asset_name="DOES_NOT_EXIST",
            )

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 3

        my_validator = context.get_validator(
            datasource_name="my_directory_datasource",
            data_connector_name="my_filesystem_data_connector",
            data_asset_name="C",
            data_connector_query={"batch_filter_parameters": {"year": "2019"}},
            create_expectation_suite_with_name="my_expectations",
            batch_spec_passthrough={
                "sampling_method": "_sample_using_hash",
                "sampling_kwargs": {
                    "column_name": "date",
                    "hash_function_name": "md5",
                    "hash_value": "f",
                },
            },
        )
        my_evr = my_validator.expect_column_values_to_be_between(
            column="d", min_value=1, max_value=31
        )
        assert my_evr.success

        # "DataContext.get_batch()" calls "DataContext.get_batch_list()" (decorated by "@usage_statistics_enabled_method").
        assert mock_emit.call_count == 4

        # my_evr = my_validator.expect_table_columns_to_match_ordered_list(ordered_list=["x", "y", "z"])
        # assert my_evr.success

        # No other usage stats calls detected
        assert mock_emit.call_count == 4

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_golden_path_runtime_data_connector_pandas_datasource_configuration(
    mock_emit, caplog, empty_data_context_stats_enabled, test_df, tmp_path_factory
):
    """
    Tests output of test_yaml_config() for a Datacontext configured with a Datasource with
    RuntimeDataConnector. Even though the test directory contains multiple files that can be read-in
    by GX, the RuntimeDataConnector will output 0 data_assets, and return a "note" to the user.

    This is because the RuntimeDataConnector is not aware of data_assets until they are passed in
    through the RuntimeBatchRequest.

    The test asserts that the proper number of data_asset_names are returned and note is returned to the user.
    """
    base_directory = str(
        tmp_path_factory.mktemp("test_golden_path_pandas_datasource_configuration")
    )

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_charlie/A/A-1.csv",
            "test_dir_charlie/A/A-2.csv",
            "test_dir_charlie/A/A-3.csv",
        ],
        file_content_fn=lambda: test_df.to_csv(header=True, index=False),
    )

    context: DataContext = empty_data_context_stats_enabled

    with set_directory(context.root_directory):
        import great_expectations as gx

        context = gx.get_context()
        mock_emit.reset_mock()  # Remove data_context.__init__ call

        yaml_config = """
           class_name: Datasource

           execution_engine:
               class_name: PandasExecutionEngine

           data_connectors:
               default_runtime_data_connector_name:
                   class_name: RuntimeDataConnector
                   batch_identifiers:
                       - default_identifier_name
           """

        # noinspection PyUnusedLocal
        report_object = context.test_yaml_config(
            name="my_directory_datasource",
            yaml_config=yaml_config,
            return_mode="report_object",
        )

        assert report_object["execution_engine"] == {
            "caching": True,
            "module_name": "great_expectations.execution_engine.pandas_execution_engine",
            "class_name": "PandasExecutionEngine",
            "discard_subset_failing_expectations": False,
            "boto3_options": {},
            "azure_options": {},
            "gcs_options": {},
        }
        assert report_object["data_connectors"]["count"] == 1

        # checking the correct number of data_assets have come back
        assert (
            report_object["data_connectors"]["default_runtime_data_connector_name"][
                "data_asset_count"
            ]
            == 0
        )

        # checking that note has come back
        assert (
            report_object["data_connectors"]["default_runtime_data_connector_name"][
                "note"
            ]
            == "RuntimeDataConnector will not have data_asset_names until they are passed in through RuntimeBatchRequest"
        )

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_golden_path_runtime_data_connector_and_inferred_data_connector_pandas_datasource_configuration(
    mock_emit, caplog, empty_data_context_stats_enabled, test_df, tmp_path_factory
):
    """
    Tests output of test_yaml_config() for a Datacontext configured with a Datasource with InferredAssetDataConnector
    and RuntimeDataConnector.

    1. The InferredAssetDataConnector will output 4 data_assets, which correspond to the files in the test_dir_charlie folder

    2.  RuntimeDataConnector will output 0 data_assets, and return a "note" to the user. This is because the
        RuntimeDataConnector is not aware of data_assets until they are passed in through the RuntimeBatchRequest.

    The test asserts that the proper number of data_asset_names are returned for both DataConnectors, and in the case of
    the RuntimeDataConnetor, the proper note is returned to the user.
    """
    base_directory = str(
        tmp_path_factory.mktemp("test_golden_path_pandas_datasource_configuration")
    )

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_charlie/A/A-1.csv",
            "test_dir_charlie/A/A-2.csv",
            "test_dir_charlie/A/A-3.csv",
            "test_dir_charlie/B/B-1.csv",
            "test_dir_charlie/B/B-2.csv",
            "test_dir_charlie/B/B-3.csv",
            "test_dir_charlie/C/C-1.csv",
            "test_dir_charlie/C/C-2.csv",
            "test_dir_charlie/C/C-3.csv",
            "test_dir_charlie/D/D-1.csv",
            "test_dir_charlie/D/D-2.csv",
            "test_dir_charlie/D/D-3.csv",
        ],
        file_content_fn=lambda: test_df.to_csv(header=True, index=False),
    )

    context: DataContext = empty_data_context_stats_enabled

    with set_directory(context.root_directory):
        import great_expectations as gx

        context = gx.get_context()
        mock_emit.reset_mock()  # Remove data_context.__init__ call

        yaml_config = f"""
        class_name: Datasource

        execution_engine:
            class_name: PandasExecutionEngine

        data_connectors:
            default_runtime_data_connector_name:
                class_name: RuntimeDataConnector
                batch_identifiers:
                    - default_identifier_name
            default_inferred_data_connector_name:
                class_name: InferredAssetFilesystemDataConnector
                base_directory: {base_directory}/test_dir_charlie
                glob_directive: "*/*.csv"

                default_regex:
                    pattern: (.+)/(.+)-(\\d+)\\.csv
                    group_names:
                        - subdirectory
                        - data_asset_name
                        - number
        """

        # noinspection PyUnusedLocal
        report_object = context.test_yaml_config(
            name="my_directory_datasource",
            yaml_config=yaml_config,
            return_mode="report_object",
        )

        assert report_object["execution_engine"] == {
            "caching": True,
            "module_name": "great_expectations.execution_engine.pandas_execution_engine",
            "class_name": "PandasExecutionEngine",
            "discard_subset_failing_expectations": False,
            "boto3_options": {},
            "azure_options": {},
            "gcs_options": {},
        }
        assert report_object["data_connectors"]["count"] == 2
        assert report_object["data_connectors"][
            "default_runtime_data_connector_name"
        ] == {
            "class_name": "RuntimeDataConnector",
            "data_asset_count": 0,
            "data_assets": {},
            "example_data_asset_names": [],
            "example_unmatched_data_references": [],
            "note": "RuntimeDataConnector will not have data_asset_names until they are "
            "passed in through RuntimeBatchRequest",
            "unmatched_data_reference_count": 0,
        }
        assert report_object["data_connectors"][
            "default_inferred_data_connector_name"
        ] == {
            "class_name": "InferredAssetFilesystemDataConnector",
            "data_asset_count": 4,
            "example_data_asset_names": ["A", "B", "C"],
            "data_assets": {
                "A": {
                    "batch_definition_count": 3,
                    "example_data_references": ["A/A-1.csv", "A/A-2.csv", "A/A-3.csv"],
                },
                "B": {
                    "batch_definition_count": 3,
                    "example_data_references": ["B/B-1.csv", "B/B-2.csv", "B/B-3.csv"],
                },
                "C": {
                    "batch_definition_count": 3,
                    "example_data_references": ["C/C-1.csv", "C/C-2.csv", "C/C-3.csv"],
                },
            },
            "unmatched_data_reference_count": 0,
            "example_unmatched_data_references": [],
        }

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_rule_based_profiler_integration(
    mock_emit, caplog, empty_data_context_stats_enabled, test_df, tmp_path_factory
):
    context = empty_data_context_stats_enabled
    yaml_config = """
    name: my_profiler
    class_name: RuleBasedProfiler
    module_name: great_expectations.rule_based_profiler
    config_version: 1.0
    variables:
      integer_type: INTEGER
      timestamp_type: TIMESTAMP
      max_user_id: 999999999999
      min_timestamp: 2004-10-19 10:23:54
    rules:
      my_rule_for_user_ids:
        domain_builder:
          class_name: TableDomainBuilder
        expectation_configuration_builders:
          - expectation_type: expect_column_values_to_be_of_type
            class_name: DefaultExpectationConfigurationBuilder
    """
    instantiated_class = context.test_yaml_config(
        yaml_config=yaml_config, name="my_profiler", class_name="Profiler"
    )

    # Ensure valid return type and content
    assert isinstance(instantiated_class, RuleBasedProfiler)
    assert instantiated_class.name == "my_profiler"

    # Confirm that logs do not contain any exceptions or invalid messages
    assert not usage_stats_exceptions_exist(messages=caplog.messages)
    assert not usage_stats_invalid_messages_exist(messages=caplog.messages)


def test_test_yaml_config_supported_types_have_self_check():
    # Each major category of test_yaml_config supported types has its own origin module_name
    supported_types = [
        (
            _YamlConfigValidator.TEST_YAML_CONFIG_SUPPORTED_STORE_TYPES,
            "great_expectations.data_context.store",
        ),
        (
            _YamlConfigValidator.TEST_YAML_CONFIG_SUPPORTED_DATASOURCE_TYPES,
            "great_expectations.datasource",
        ),
        (
            _YamlConfigValidator.TEST_YAML_CONFIG_SUPPORTED_DATA_CONNECTOR_TYPES,
            "great_expectations.datasource.data_connector",
        ),
        (
            _YamlConfigValidator.TEST_YAML_CONFIG_SUPPORTED_CHECKPOINT_TYPES,
            "great_expectations.checkpoint",
        ),
        (
            _YamlConfigValidator.TEST_YAML_CONFIG_SUPPORTED_PROFILER_TYPES,
            "great_expectations.rule_based_profiler",
        ),
    ]

    # Quick sanity check to ensure that we are testing ALL supported types herein
    all_types = list(itertools.chain.from_iterable(t[0] for t in supported_types))
    assert sorted(all_types) == sorted(
        _YamlConfigValidator.ALL_TEST_YAML_CONFIG_SUPPORTED_TYPES
    )

    # Use class_name and module_name to get the class type and introspect to confirm adherence to self_check requirement
    for category, module_name in supported_types:
        for class_name in category:
            class_ = load_class(class_name=class_name, module_name=module_name)
            assert hasattr(class_, "self_check") and callable(
                class_.self_check
            ), f"Class '{class_}' is missing the required `self_check()` method"


@pytest.mark.integration
def test_test_yaml_config_on_datasource_sanitizes_instantiated_objs_config(
    empty_data_context_stats_enabled,
    monkeypatch,
):
    context = empty_data_context_stats_enabled
    validator = _YamlConfigValidator(context)

    variable = "DATA_DIR"
    value_associated_with_variable = "a/b/c"
    data_connector_name = "my_data_connector"

    monkeypatch.setenv(variable, value_associated_with_variable)

    yaml_config = f"""
name: my_datasource
class_name: Datasource
execution_engine:
  class_name: PandasExecutionEngine
data_connectors:
  {data_connector_name}:
    class_name: InferredAssetFilesystemDataConnector
    base_directory: ${variable}
"""
    instantiated_class = validator.test_yaml_config(yaml_config=yaml_config)

    # Runtime object should have the substituted value for downstream usage
    assert instantiated_class.data_connectors[
        data_connector_name
    ].base_directory.endswith(value_associated_with_variable)

    # Config attached to object should mirror the runtime object
    assert instantiated_class.config["data_connectors"][data_connector_name][
        "base_directory"
    ].endswith(value_associated_with_variable)

    # Raw config attached to object should reflect what needs to be persisted (no sensitive credentials!)
    assert (
        instantiated_class._raw_config["data_connectors"][data_connector_name][
            "base_directory"
        ]
        == f"${variable}"
    )



import pytest
import requests
from requests.adapters import HTTPAdapter, Retry


@pytest.fixture(scope="session")
def requests_session_with_retries() -> requests.Session:
    # https://stackoverflow.com/a/35636367
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session



import json
from typing import Optional, Tuple

from great_expectations.core._docs_decorators import public_api
from great_expectations.core.domain import Domain
from great_expectations.core.id_dict import IDDict
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.util import convert_to_json_serializable


@public_api
class MetricConfiguration:
    """An interface for configuring Metrics.

    MetricConfiguration allows the configuration of domain information, dependencies and additional metric-specific
    configurations.  Metrics are computed attributes of data, and are derived from one or more Batches that can then
    be used to evaluate Expectations or to summarize the result of the Validation.

    Args:
        metric_name (str): name of the Metric defined by the current MetricConfiguration.
        metric_domain_kwargs (dict): provides information on where the Metric can be calculated. For instance, a
            MapCondition metric can include the name of the column that the Metric is going to be run on.
        metric_value_kwargs (optional[dict]): Optional kwargs that define values specific to each Metric.  For instance,
            a Metric that partitions a column can define the method of partitioning (`uniform` bins) and the number of
            bins (`n_bins`) as `metric_value_kwargs`.
        metric_dependencies (optional[dict]): This is a dict consisting of all Metrics necessary to evaluate the Expectation.
    """

    def __init__(
        self,
        metric_name: str,
        metric_domain_kwargs: dict,
        metric_value_kwargs: Optional[dict] = None,
        metric_dependencies: Optional[dict] = None,
    ) -> None:
        self._metric_name = metric_name

        if not isinstance(metric_domain_kwargs, IDDict):
            metric_domain_kwargs = IDDict(metric_domain_kwargs)

        self._metric_domain_kwargs: IDDict = metric_domain_kwargs

        if not isinstance(metric_value_kwargs, IDDict):
            if metric_value_kwargs is None:
                metric_value_kwargs = {}
            metric_value_kwargs = IDDict(metric_value_kwargs)

        self._metric_value_kwargs: IDDict = metric_value_kwargs

        self._metric_dependencies: IDDict = IDDict({})
        if metric_dependencies is not None:
            self._metric_dependencies = IDDict(metric_dependencies)

    def __repr__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def __str__(self):
        return self.__repr__()

    @property
    def metric_name(self) -> str:
        return self._metric_name

    @property
    def metric_domain_kwargs(self) -> IDDict:
        return self._metric_domain_kwargs

    @property
    def metric_value_kwargs(self) -> IDDict:
        return self._metric_value_kwargs

    @property
    def metric_domain_kwargs_id(self) -> str:
        return self.metric_domain_kwargs.to_id()

    @property
    def metric_value_kwargs_id(self) -> str:
        return self.metric_value_kwargs.to_id()

    @property
    def metric_dependencies(self) -> IDDict:
        return self._metric_dependencies

    @metric_dependencies.setter
    def metric_dependencies(self, metric_dependencies) -> None:
        self._metric_dependencies = metric_dependencies

    def get_domain(self) -> Domain:
        """Return "Domain" object, constructed from this "MetricConfiguration" object."""
        domain_type: MetricDomainTypes = self.get_domain_type()

        if domain_type == MetricDomainTypes.TABLE:
            other_table_name: Optional[str] = self._metric_domain_kwargs.get("table")
            if other_table_name:
                return Domain(
                    domain_type=domain_type,
                    domain_kwargs={
                        "table": other_table_name,
                    },
                )

            return Domain(
                domain_type=domain_type,
            )

        if domain_type == MetricDomainTypes.COLUMN:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column": self._metric_domain_kwargs["column"],
                },
            )

        if domain_type == MetricDomainTypes.COLUMN_PAIR:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column_A": self._metric_domain_kwargs["column_A"],
                    "column_B": self._metric_domain_kwargs["column_B"],
                },
            )

        if domain_type == MetricDomainTypes.MULTICOLUMN:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column_list": self._metric_domain_kwargs["column_list"],
                },
            )

        raise ValueError(f"""Domain type "{domain_type}" is not recognized.""")

    def get_domain_type(self) -> MetricDomainTypes:
        """Return "domain_type" of this "MetricConfiguration" object."""
        if "column" in self._metric_domain_kwargs:
            return MetricDomainTypes.COLUMN

        if (
            "column_A" in self._metric_domain_kwargs
            and "column_B" in self._metric_domain_kwargs
        ):
            return MetricDomainTypes.COLUMN_PAIR

        if "column_list" in self._metric_domain_kwargs:
            return MetricDomainTypes.MULTICOLUMN

        # TODO: <Alex>Determining "domain_type" of "MetricConfiguration" using heuristics defaults to "TABLE".</Alex>
        return MetricDomainTypes.TABLE

    @property
    def id(self) -> Tuple[str, str, str]:
        return (
            self.metric_name,
            self.metric_domain_kwargs_id,
            self.metric_value_kwargs_id,
        )

    @public_api
    def to_json_dict(self) -> dict:
        """Returns a JSON-serializable dict representation of this MetricConfiguration.

        Returns:
            A JSON-serializable dict representation of this MetricConfiguration.

        """
        json_dict: dict = convert_to_json_serializable(
            data={
                "metric_name": self.metric_name,
                "metric_domain_kwargs": self.metric_domain_kwargs,
                "metric_domain_kwargs_id": self.metric_domain_kwargs_id,
                "metric_value_kwargs": self.metric_value_kwargs,
                "metric_value_kwargs_id": self.metric_value_kwargs_id,
                "id": self.id,
            }
        )
        return json_dict



from unittest import mock

from great_expectations.cli.batch_request import (
    _get_data_asset_name_from_data_connector,
)


@mock.patch("great_expectations.cli.batch_request.BaseDatasource")
@mock.patch("great_expectations.cli.batch_request._get_user_response")
def test_get_data_asset_name_from_data_connector_default_path(
    mock_user_input, mock_datasource
):
    mock_datasource.get_available_data_asset_names.return_value = {
        "my_data_connector": ["a", "b", "c", "d", "e"]
    }
    mock_user_input.side_effect = ["4"]  # Immediately select my asset
    data_asset_name = _get_data_asset_name_from_data_connector(
        mock_datasource, "my_data_connector", "my message prompt"
    )
    assert data_asset_name == "d"


@mock.patch("great_expectations.cli.batch_request.BaseDatasource")
@mock.patch("great_expectations.cli.batch_request._get_user_response")
def test_get_data_asset_name_from_data_connector_pagination(
    mock_user_input, mock_datasource
):
    mock_datasource.get_available_data_asset_names.return_value = {
        "my_data_connector": [f"my_file{n}" for n in range(200)]
    }
    mock_user_input.side_effect = [
        "l",  # Select listing/pagination option
        "n",  # Go to page 2 of my data asset listing
        "n",  # Go to page 3 of my data asset listing
        "34",  # Select the 34th option in page 3
    ]
    data_asset_name = _get_data_asset_name_from_data_connector(
        mock_datasource, "my_data_connector", "my message prompt"
    )
    assert data_asset_name == "my_file128"


@mock.patch("great_expectations.cli.batch_request.BaseDatasource")
@mock.patch("great_expectations.cli.batch_request._get_user_response")
def test_get_data_asset_name_from_data_connector_with_search(
    mock_user_input, mock_datasource
):
    files = [f"my_file{n}" for n in range(200)]
    target_file = "my_file2021-12-30"
    files.append(target_file)
    mock_datasource.get_available_data_asset_names.return_value = {
        "my_data_connector": files
    }
    mock_user_input.side_effect = [
        "s",  # Select search option
        "my_file20",  # Filter listing
        r"my_file\d{4}-\d{2}-\d{2}",  # Use regex to isolate one file with date format
        "1",  # Select the 1st and only option
    ]
    data_asset_name = _get_data_asset_name_from_data_connector(
        mock_datasource, "my_data_connector", "my message prompt"
    )
    assert data_asset_name == target_file
